#!/usr/bin/env python
#
# frame.py - A wx.Frame which implements a 3D image viewer.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
# Author: Taylor Hanayik <taylor.hanayik@ndcn.ox.ac.uk>
#
"""This module provides the :class:`FSLeyesFrame` which is the top level frame
for FSLeyes.
"""


from __future__ import division

import functools as ft
import itertools as it
import              logging

from typing import Type

import wx
import wx.lib.agw.aui               as aui

import fsl.utils.idle               as idle
import fsl.utils.settings           as fslsettings
import fsleyes_widgets              as fwidgets
import fsleyes_widgets.dialog       as fsldlg
import fsleyes_widgets.utils.status as status

import fsleyes.strings              as strings
import fsleyes.plugins              as plugins
import fsleyes.autodisplay          as autodisplay
import fsleyes.profiles.shortcuts   as shortcuts
import fsleyes.views.viewpanel      as viewpanel
import fsleyes.actions              as actions
import fsleyes.tooltips             as tooltips
import fsleyes.layouts              as layouts
import fsleyes.displaycontext       as displaycontext


log = logging.getLogger(__name__)


class FSLeyesFrame(wx.Frame):
    """A ``FSLeyesFrame`` is a simple :class:`wx.Frame` which acts as a
    container for :class:`.ViewPanel` instances.


    A :class:`wx.lib.agw.aui.AuiManager` is used so that ``ViewPanel`` panels
    can be dynamically laid out and reconfigured by the user.


    **Menus**


    The ``FSLeyesFrame`` has the following menus:

    ========== ==============================================================
    *FSLeyes*  General actions, such as help, quit, about.
    *File*     File actions such as adding a new overlay.
    *Overlay*  Actions for working with overlays.
    *View*     Options to open a new :class:`.ViewPanel`.
    *Settings* Options which are specific to the currently visible
               :class:`.ViewPanel` instances. A separate sub-menu is added
               for each visible ``ViewPanel``. All ``ViewPanel`` classes
               inherit from the :class:`.ActionProvider` class - any actions
               that have been defined are added as menu items here.
    *Tools*    Options which invoke some sort of tool - some global tools are
               added to this menu, along with tools that are specific to each
               active view.
    ========== ==============================================================


    **Saving/restoring state**


    When a ``FSLeyesFrame`` is closed, it saves some display settings so that
    they can be restored the next time a ``FSLeyesFrame`` is opened. The
    settings are saved using the :class:`~fsl.utils.settings` module.
    Currently, the frame position, size, and layout (see the
    :mod:`.layout` module) are saved.


    **Programming interface**


    The ``FSLeyesFrame`` provides the following properties and methods for
    programmatically configuring the display:

    .. autosummary::
       :nosignatures:

       overlayList
       displayCtx
       viewPanels
       focusedViewPanel
       auiManager
       getViewPanelInfo
       getViewPanelID
       getViewPanelTitle
       addViewPanel
       viewPanelDefaultLayout
       removeViewPanel
       removeAllViewPanels
       refreshViewMenu
       refreshLayoutMenu
       refreshSettingsMenu
       refreshToolsMenu
       populateMenu
       Close


    **Actions**


    The :mod:`fsleyes.actions.frameactions` module contains some
    :mod:`.actions` which are monkey-patched into the ``FSLeyesFrame`` class.
    These actions are made available to the user via menu items and/or keyboard
    shortcuts.


    .. note:: All of the functions defined in the
              :mod:`fsleyes.actions.frameactions` module are treated as
              first-class methods of the ``FSLeyesFrame`` class (i.e. they are
              assumed to be present). They are only in a separate module to
              keep file sizes down.
    """

    def __init__(self,
                 parent,
                 overlayList,
                 displayCtx,
                 restore=False,
                 save=True,
                 fontSize=None,
                 menu=True,
                 closeHandlers=None):
        """Create a ``FSLeyesFrame``.

        :arg parent:        The :mod:`wx` parent object.

        :arg overlayList:   The :class:`.OverlayList`.

        :arg displayCtx:    The master :class:`.DisplayContext`.

        :arg restore:       Restores previous saved layout. If ``False``, no
                            view panels will be displayed.

        :arg save:          Save current layout when closed.

        :arg fontSize:      Application-wide font size to use. Defaults to 10.

        :arg menu:          Whether or not to create a menu bar.

        :arg closeHandlers: List of functions to call when this
                            ``FSLeyesFrame`` is closing. Use this rather than
                            binding to the ``EVT_CLOSE`` event, so that the
                            ``FSLeyesFrame`` can guarantee that your handler
                            will be called when it is actually closing.
                            Otherwise the user might cancel the close, but
                            your handler will still get called.
        """
        wx.Frame.__init__(self, parent, title='FSLeyes')
        tooltips.initTooltips()

        # Default application font - this is
        # inherited by all child controls.
        font = self.GetFont()

        if fontSize is None:
            fontSize = 10

        font.SetPointSize(fontSize)
        font.SetWeight(wx.FONTWEIGHT_NORMAL)
        self.SetFont(font)

        self.__overlayList = overlayList
        self.__displayCtx  = displayCtx
        self.__name        = '{}_{}'.format(type(self).__name__, id(self))
        self.__mainPanel   = wx.Panel(self)
        self.__statusBar   = wx.StaticText(self)

        # Even though the FSLeyesFrame does not allow
        # panels to be floated, I am specifying the
        # docking guide style for complicated reasons...
        #
        # Each ViewPanel contained in this FSLeyesFrame
        # has an AuiManager of its own; these child
        # AuiManagers do support floating of their
        # child panels. However, it seems that when
        # a floating child panel of a ViewPanel is
        # docked, this top-level AuiManager is called
        # to draw the docking guides. This is because
        # the wx.lib.agw.aui.framemanager.GetManager
        # function uses the wx event handling system
        # to figure out which AuiManager should be used
        # to manage the docking (which is a ridiculous
        # way to do this, in my opinion).
        #
        # Anyway, this means that the docking guides
        # will be drawn according to the style set up
        # in this AuiManager, instead of the ViewPanel
        # AuiManager, which is the one that is actually
        # managing the panel being docked.
        #
        # This wouldn't be a problem, if not for the fact
        # that, when running over SSH/X11, the default
        # docking guides seem to get sized incorrectly,
        # and look terrible (probably related to the
        # AuiDockingGuide monkey-patch at the bottom of
        # viewpanel.py).
        #
        # This problem does not occur with the aero/
        # whidbey guides.
        self.__auiManager  = aui.AuiManager(
            self.__mainPanel,
            agwFlags=(aui.AUI_MGR_RECTANGLE_HINT          |
                      aui.AUI_MGR_NO_VENETIAN_BLINDS_FADE |
                      aui.AUI_MGR_AERO_DOCKING_GUIDES     |
                      aui.AUI_MGR_LIVE_RESIZE))

        self.__auiManager.SetDockSizeConstraint(0.5, 0.5)

        self.__sizer = wx.BoxSizer(wx.VERTICAL)
        self.__sizer.Add(self.__mainPanel, flag=wx.EXPAND, proportion=1)
        self.__sizer.Add(self.__statusBar, flag=wx.EXPAND)

        self.SetSizer(self.__sizer)

        # Re-direct status updates
        # to the status bar.
        def update(msg):

            # Do the update via wx.CallAfter to
            # make sure that the status bar is
            # updated on the main loop.
            def realUpdate():

                # This function might get called after
                # the status bar has been destroyed,
                # so we'll absorb any errors.
                try:
                    self.__statusBar.SetLabel(msg)
                    self.__statusBar.Refresh()
                    self.__statusBar.Update()
                except Exception:
                    pass

            wx.CallAfter(realUpdate)
        status.setTarget(update)

        # Keeping track of all open view panels
        #
        # The __viewPanels list contains all
        # [ViewPanel] instances
        #
        # The other dicts contain
        # {ViewPanel : something} mappings
        #
        self.__viewPanels      = []
        self.__viewPanelDCs    = {}
        self.__viewPanelMenus  = {}
        self.__viewPanelIDs    = {}
        self.__viewPanelTitles = {}

        # Refs to menus
        self.__haveMenu       = menu
        self.__menuBar        = None
        self.__fsleyesMenu    = None
        self.__overlayMenu    = None
        self.__fileMenu       = None
        self.__viewMenu       = None
        self.__layoutMenu     = None
        self.__settingsMenu   = None
        self.__toolsMenu      = None
        self.__recentPathMenu = None

        # Refs to Action objects that are managed
        # by this FSLeyesFrame, and are bound to
        # menu options, as { type : instance }
        # mappings. This includes items in the
        # File, Overlay, View, and Tools menus
        self.__menuActions = {}

        # We keep refs to the (Action, wx.MenuItem)
        # pairs for each menu, because when a menu
        # gets destroyed, we need to unbind the action
        # from the item
        self.__viewMenuActions      = []
        self.__viewPanelMenuActions = {}
        self.__layoutMenuActions    = []
        self.__overlayMenuActions   = []
        self.__toolsMenuActions     = []

        # The recent paths manager notifies us when
        # they change. See the __makeFileMenu and
        # __makeRecentPathsMenu methods
        import fsleyes.actions.loadoverlay as loadoverlay
        loadoverlay.recentPathManager.register(
            self.__name, self.__makeRecentPathsMenu)

        # This dictionary contains mappings of the form
        #
        #   { keyboard-shortcut : { ViewPanel : actionName } }
        #
        # See the __onViewPanelMenuItem method for details.
        self.__viewPanelShortcuts = {}

        self.__makeMenuBar()
        self.__restoreState(restore)

        # These flags control whether the user is
        # prompted before FSLeyes closes - they are
        # used in the Close and __onClose methods
        self.__saveLayout = save
        self.__askUnsaved = True

        self.__auiManager.Bind(aui.EVT_AUI_PANE_CLOSE, self.__onViewPanelClose)

        # Event handlers are called in order of
        # last bound first, first bound last.
        # Our __onClose handler may veto the
        # close (if the user cancels it), so
        # we have to bind our handler last.
        if closeHandlers is not None:
            for h in closeHandlers:
                self.Bind(wx.EVT_CLOSE, h)

        self.Bind(wx.EVT_CLOSE, self.__onClose)

        overlayList.addListener('overlays',
                                self.__name,
                                self.__selectedOverlayChanged)
        displayCtx .addListener('selectedOverlay',
                                self.__name,
                                self.__selectedOverlayChanged)

        self.Layout()


    @property
    def overlayList(self):
        """Returns the :class:`.OverlayList` which contains the overlays
        being displayed by this ``FSLeyesFrame``.
        """
        return self.__overlayList


    @property
    def displayCtx(self):
        """Returns the top-level :class:`.DisplayContext` associated with this
        ``FSLeyesFrame``.
        """
        return self.__displayCtx


    @property
    def viewPanels(self):
        """Returns a list of all :class:`.ViewPanel` instances that are
        currenlty displayed in this ``FSLeyesFrame``.
        """
        return list(self.__viewPanels)


    def getView(self, panelType):
        """Return a list of all :class:`.ViewPanel` instances of the specified
        type.
        """
        return [v for v in self.__viewPanels if isinstance(v, panelType)]


    @property
    def focusedViewPanel(self):
        """Returns the :class:`.ViewPanel` which currently has focus,
        the first ``ViewPanel`` if none have focus, or ``None`` if there
        are no ``ViewPanels``.
        """

        if len(self.viewPanels) == 0:
            return None

        focused = wx.Window.FindFocus()

        while focused is not None:

            if isinstance(focused, viewpanel.ViewPanel):
                return focused

            focused = focused.GetParent()

        return self.viewPanels[0]


    @property
    def menuActions(self):
        """Returns a dictionary containing all global :class:`.Action` objects
        that are bound to menu items.
        """
        return dict(self.__menuActions)


    def getViewPanelID(self, viewPanel):
        """Returns the ID that was assigned to the given :class:`.ViewPanel`.
        This is a sequentially increasing integer, starting from 1.
        """
        return self.__viewPanelIDs[viewPanel]


    def getViewPanelTitle(self, viewPanel):
        """Returns the ID that was assigned to the given :class:`.ViewPanel`.
        This is a sequentially increasing integer, starting from 1.
        """
        return self.__viewPanelTitles[viewPanel]


    def getViewPanelInfo(self, viewPanel):
        """Returns the ``AuiPaneInfo`` class which contains layout information
        about the given :class:`.ViewPanel`.
        """
        return self.__auiManager.GetPane(viewPanel)


    @property
    def auiManager(self):
        """Returns the ``wx.lib.agw.aui.AuiManager` object which is managing
        the layout of this ``FSLeyesFrame``.
        """
        return self.__auiManager


    def removeViewPanel(self, viewPanel):
        """Removes the given :class:`.ViewPanel` from this ``FSLeyesFrame``.
        """

        paneInfo = self.__auiManager.GetPane(viewPanel)

        self.__onViewPanelClose(panel=viewPanel)

        self.__auiManager.ClosePane(paneInfo)
        self.__auiManager.Update()


    def removeAllViewPanels(self):
        """Removes all view panels from this ``FSLeyesFrame``.

        .. note:: This method should be used to clear the frame, rather than
                  removing each :class:`.ViewPanel` individually via
                  :meth:`removeViewPanel`. This is because when one
                  ``ViewPanel`` is closed, the display settings of the
                  remaining ``ViewPanel`` instances may be modified.  If these
                  remaining ``ViewPanels`` are then immediately closed, thay
                  may not have had enough time to reconfigure themselves (e.g.
                  :class:`.GLVolume` instances re-creating their
                  :class:`.ImageTexture` instance due to a change in
                  :attr:`.DisplayContext.syncOverlayDisplay`), and ugly things
                  will happen (e.g. an :class:`.ImageTexture` trying to
                  configure itself *after* it has already been destroyed).

                  So just use this method instead.
        """
        for vp in list(self.__viewPanels):

            paneInfo = self.__auiManager.GetPane(vp)
            self.__onViewPanelClose(panel=vp, displaySync=False)
            self.__auiManager.ClosePane(paneInfo)

        self.__auiManager.Update()


    def addViewPanel(
            self,
            panelCls      : Type[viewpanel.ViewPanel],
            title         : str  = None,
            defaultLayout : bool = True,
            **kwargs) -> viewpanel.ViewPanel:
        """Adds a new :class:`.ViewPanel` to the centre of the frame, and a
        menu item allowing the user to configure the view.

        :arg panelCls:      The :class:`.ViewPanel` type to be added.

        :arg title:         Title to give the view. If not provided, it is
                            assumed that a name is present for the view type
                            in :attr:`.strings.titles`.

        :arg defaultLayout: If ``True`` (the default),
                            :meth:`viewPanelDefaultLayout` is called, to add
                            a default set of control panels to the view.

        :returns: The newly created ``ViewPanel``.

        All other arguments are passed to the ``__init__`` method of the child
        :class:`.DisplayContext` that is created for the new view.
        """
        import fsleyes.views.plotpanel  as plotpanel
        import fsleyes.views.shellpanel as shellpanel

        if title is None:
            title = panelCls.title()
        if title is None:
            title = strings.titles.get(panelCls, panelCls.__name__)

        if len(self.__viewPanelIDs) == 0:
            panelId = 1
        else:
            panelId = max(self.__viewPanelIDs.values()) + 1

        # The PaneInfo Name contains the panel
        # class name - this is used for saving
        # and restoring layouts .
        name  = '{} {}'.format(panelCls.__name__, panelId)
        title = '{} {}'.format(title,             panelId)

        childDC = displaycontext.DisplayContext(
            self.__overlayList,
            parent=self.__displayCtx,
            **kwargs)

        panel = panelCls(self.__mainPanel, self.__overlayList, childDC, self)

        log.debug('Created new {} ({}) with DisplayContext {}'.format(
            panelCls.__name__,
            id(panel),
            id(childDC)))

        paneInfo = (aui.AuiPaneInfo()
                    .Name(name)
                    .Caption(title)
                    .CloseButton()
                    .Dockable()
                    .Resizable()
                    .DestroyOnClose())

        # When there is only one view panel
        # displayed, the AuiManager seems to
        # have trouble drawing the caption
        # bar - it is drawn, but then the
        # panel is drawn over the top of it.
        # So if we only have one panel, we
        # hide the caption bar
        if panelId == 1:
            paneInfo.Centre().Dockable(False).CaptionVisible(False)

        # But then re-show it when another
        # panel is added. The __viewPanels
        # dict is an OrderedDict, so the
        # first key is the AuiPaneInfo of
        # the first panel that was added.
        else:
            self.__auiManager.GetPane(self.__viewPanels[0])\
                             .CaptionVisible(True)

        # If this is not the first view panel,
        # give it a sensible initial size.
        if panelId > 1:
            width, height = self.GetClientSize().Get()

            # PlotPanels are initially
            # placed along the bottom
            if isinstance(panel, plotpanel.PlotPanel):
                paneInfo.Bottom().BestSize(width, height // 3)

            # As are ShellPanels, albeit
            # a bit smaller
            elif isinstance(panel, shellpanel.ShellPanel):
                paneInfo.Bottom().BestSize(width, height // 5)

            # Other panels (e.g. CanvasPanels)
            # are placed on the right
            else:
                paneInfo.Right().BestSize(width // 2, height)

        self.__viewPanels.append(panel)
        self.__viewPanelDCs[     panel] = childDC
        self.__viewPanelIDs[     panel] = panelId
        self.__viewPanelTitles[  panel] = title

        menu, actionItems = self.__addViewPanelMenu(  panel, title)
        self.__viewPanelMenus[      panel] = menu
        self.__viewPanelMenuActions[panel] = actionItems

        self.__auiManager.AddPane(panel, paneInfo)
        self.__configDisplaySync(panel)
        self.__auiManager.Update()
        self.__enableMenus(True)

        if defaultLayout:
            self.viewPanelDefaultLayout(panel)

        return panel


    def viewPanelDefaultLayout(self, viewPanel : viewpanel.ViewPanel):
        """After a :class:`.ViewPanel` is added via the view menu, this method
        is called to perform some basic initialisation on the panel. This
        basically amounts to adding toolbars.
        """

        viewPanel.removeAllPanels()
        ctrls = viewPanel.defaultLayout()
        if ctrls is None:
            return

        ctrls = [plugins.lookupControl(c) for c in ctrls]

        for ctrl in ctrls:
            title = plugins.pluginTitle(ctrl)
            viewPanel.togglePanel(ctrl, title=title)


    def refreshViewMenu(self):
        """Re-creates the *View* menu."""

        if not self.__haveMenu:
            return

        viewMenu = self.__viewMenu
        items    = self.__viewMenuActions

        for action, item in items:
            action.unbindWidget(item)

        for item in viewMenu.GetMenuItems():
            viewMenu.Delete(item.GetId())

        self.__viewMenuActions = self.__makeViewMenu()
        self.refreshLayoutMenu()


    def refreshLayoutMenu(self):
        """Re-creates the *View -> Layouts* sub-menu. """

        if not self.__haveMenu:
            return

        for action, item in self.__layoutMenuActions:
            action.unbindWidget(item)
            action.destroy()

        for item in self.__layoutMenu.GetMenuItems():
            self.__layoutMenu.Delete(item.GetId())

        self.__layoutMenuActions = self.__makeLayoutMenu()


    def refreshSettingsMenu(self):
        """Re-creates the *Settings* menu. """

        if not self.__haveMenu:
            return

        for action, item in it.chain(*self.__viewPanelMenuActions.values()):
            action.unbindWidget(item)

        for menu in self.__viewPanelMenus.values():
            menu = menu.GetSubMenu()
            for item in menu.GetMenuItems():
                menu.Delete(item.GetId())

        for item in self.__settingsMenu.GetMenuItems():
            self.__settingsMenu.Delete(item.GetId())

        self.__viewPanelMenus       = {}
        self.__viewPanelMenuActions = {}
        for panel, title in self.__viewPanelTitles.items():

            menu, actionItems = self.__addViewPanelMenu(panel, title)

            self.__viewPanelMenus[      panel] = menu
            self.__viewPanelMenuActions[panel] = actionItems


    def refreshToolsMenu(self):
        """Re-creates the *Tools* menu. """

        if not self.__haveMenu:
            return

        menu  = self.__toolsMenu
        items = self.__toolsMenuActions

        for action, item in items:
            action.unbindWidget(item)

        # all items from self.__toolMenuActions
        # will also be returned by GetMenuItems,
        # along with other things (e.g. section
        # dividers)
        for item in menu.GetMenuItems():
            menu.Delete(item.GetId())

        self.__toolsMenuActions = self.__makeToolsMenu()


    def populateMenu(self,
                     menu,
                     target,
                     actionNames=None,
                     actionTitles=None,
                     **kwargs):
        """Creates menu items for every :class:`.Action` available on the
        given ``target``, or for every named action in the ``actionNames``
        list.

        Called by the :meth:`__addViewPanelMenu` method to generate a menu
        for new :class:`.ViewPanel` instances, but can also be called for
        other purposes.

        :arg menu:         The ``wx.Menu`` to be populated.

        :arg target:       The object which has actions to be bound to the
                           menu items.

        :arg actionNames:  If provided, only menu items for the actions named
                           in this list will be created. May contain ``None``,
                           which indicates that a menu separator should be
                           added at that point.

        :arg actionTitles: Optional dict containing ``{name : title}`` mappings
                           for some actions. If not provided, it is assumed
                           that a name for the action exists in
                           :attr:`.strings.actions`.

        All other keyword arguments are passed through to the
        :meth:`__onViewPanelMenuItem` method.

        :returns: A list containing the ``(Action, wx.MenuItem)`` pairs that
                  were added to the menu.
        """
        if actionTitles is None:
            actionTitles = {}

        if actionNames is None:
            actionNames, actionObjs = list(zip(*target.getActions()))
        else:
            actionObjs = [target.getAction(name)
                          if name is not None else None
                          for name in actionNames]

        actionNames  = list(actionNames)
        actionObjs   = list(actionObjs)
        actionTitles = [actionTitles.get(
            n, strings.actions.get((target, n), n)) for n in actionNames]

        def configureActionItem(menu, actionName, actionObj, title):

            shortcut = shortcuts.actions.get((target, actionName))

            if shortcut is not None:
                title = '{}\t{}'.format(title, shortcut)

            if isinstance(actionObj, actions.ToggleAction):
                itemType = wx.ITEM_CHECK
            else:
                itemType = wx.ITEM_NORMAL

            menuItem = menu.Append(wx.ID_ANY, title, kind=itemType)

            # If this action can be called from a
            # keyboard shortcut, we save it in a
            # dictionary for reasons explained
            # in the __onViewPanelMenuItem method.
            #
            # We formalise the shortcut string to the
            # wx representation, so it is consistent
            # regardless of whatever we have put in
            # the fsleyes.profiles.shortcuts module.
            if shortcut is not None:

                # If a shortcut uses characters
                # which are not in th platform
                # language, wx.MenuItem.GetAccel
                # returns None.
                accel = menuItem.GetAccel()

                if accel is not None:
                    shortcut     = menuItem.GetAccel().ToString()
                    shortcutList = self.__viewPanelShortcuts.get(shortcut, {})

                    shortcutList[target] = actionName

                    self.__viewPanelShortcuts[shortcut] = shortcutList

            # The __onViewPanelMenuItem method
            # needs to know which view panel/action
            # is associated with each callback.
            def wrapper(ev, vp=target, aname=actionName, sc=shortcut):
                self.__onViewPanelMenuItem(vp, aname, sc, **kwargs)

            actionObj.bindToWidget(self, wx.EVT_MENU, menuItem, wrapper)

            return menuItem

        actionItems = []
        for actionName, actionObj, actionTitle in zip(
                actionNames, actionObjs, actionTitles):

            # If actionObj is None, this is a
            # hacky hint to insert a separator
            # - see ActionProvider.getActions.
            if actionObj is None:
                menu.AppendSeparator()
                continue

            # A single normal action <-> menu item
            else:
                currentMenu = menu
                names       = [actionName]
                objs        = [actionObj]
                titles      = [actionTitle]

            for name, obj, title in zip(names, objs, titles):
                actionItems.append(
                    (obj, configureActionItem(currentMenu, name, obj, title)))

        return actionItems


    def __addViewPanelMenu(self, panel, title):
        """Called by :meth:`addViewPanel`. Adds a menu item for the newly
        created :class:`.ViewPanel` instance, and adds any tools
        for the view panel to the tools menu.

        :arg panel: The newly created ``ViewPanel`` instance.
        :arg title: The name given to the ``panel``.
        """

        if not self.__haveMenu:
            return None, []

        # The settings menu for a view is a list of its
        # actions, followed by a list of supported control
        # types. followed by a "removeAllPanels" action
        actionItems  = []
        actionNames  = [name for (name, obj) in panel.getActions()]
        actionTitles = {}
        pluginCtrls  = plugins.listControls(type(panel))

        if len(pluginCtrls) > 0:
            actionNames.append(None)

            # ViewPanel.controlOrder can suggest an ordering
            # of the control panels in the settings menu
            ctrlOrder = panel.controlOrder()
            if ctrlOrder is not None:
                names, clss = zip(*pluginCtrls.items())
                ctrlOrder   = [plugins.lookupControl(c) for c in ctrlOrder]
                indices     = [ctrlOrder.index(c) if c in ctrlOrder
                               else len(pluginCtrls) for c in clss]
                pluginCtrls = sorted(zip(indices, names, clss))
                pluginCtrls = {t[1] : t[2] for t in pluginCtrls}

            # ViewPanels have a ToggleControlPanelAction added as
            # an attributee for every supported control panel
            for ctrlName, ctrlType in pluginCtrls.items():
                name = ctrlType.__name__
                actionNames.append(name)
                actionTitles[name] = ctrlName

            # add a "remove all panels" item
            actionNames.append('removeAllPanels')

        if len(actionNames) == 0:
            return None, []

        menu    = wx.Menu()
        submenu = self.__settingsMenu.AppendSubMenu(menu, title)

        # We add a 'Close' action to the
        # menu for every panel, but put
        # another separator before it
        actionNames.append(None)
        actionNames.append('removeFromFrame')

        # Most of the work is
        # done in populateMenu
        actionItems.extend(self.populateMenu(
            menu,
            panel,
            actionNames=actionNames,
            actionTitles=actionTitles))

        # Make sure the tools
        # menu is up to date
        self.refreshToolsMenu()
        return submenu, actionItems


    def __onViewPanelMenuItem(self,
                              target,
                              actionName,
                              shortcut,
                              ignoreFocus=False,
                              runOnIdle=False):
        """Called when a menu item from a :class:`.ViewPanel` menu, or a menu
        otherwise created via :meth:`createMenu`, is selected, either via menu
        selection, or from a bound keyboard shortcut. This callback is
        configured in the :meth:`createMenu` method.

        :arg target:      The target instance for the action, most often a
                          :class:`.ViewPanel`.

        :arg actionName:  The name of the :class:`.Action` which is associated
                          with the menu item.

        :arg shortcut:    The keyboard shortcut code (see
                          ``wx.AcceleratorEntry.ToString``) associated with the
                          menu item, of ``None`` if there is no shortcut.

        :arg ignoreFocus: If ``True``, the action is executed on the
                          ``target``. Otherwise (the default), the action
                          is executed on the currently focused
                          :class:`.ViewPanel`.

        :arg runOnIdle:   If ``True``, the action is executed on the
                          ``idle.idle`` loop. Otherwise (the default),
                          the action is executed on the calling thread.
        """

        # Hacky way to see if this menu item
        # callback was triggered via a keyboard
        # shortcut or via direct menu item
        # selection.
        keyDown = any((wx.GetKeyState(wx.WXK_CONTROL),
                       wx.GetKeyState(wx.WXK_ALT),
                       wx.GetKeyState(wx.WXK_SHIFT)))

        # If there is no keyboard shortcut associated
        # with this action, or the menu item was selected
        # directly, we just execute the action directly
        # on the target which is associated with the
        # menu.
        if ignoreFocus or shortcut is None or (not keyDown):
            func = target.getAction(actionName)
            if runOnIdle:
                idle.idle(func)
            else:
                func()
            return

        # Otherwise we assume that the menu item was
        # triggered via a keyboard shortcut. In this case,
        # we want the currently focused ViewPanel to be
        # the receiver of the action callback.
        #
        # However, the same keyboard shortcut might be
        # used for different ViewPanel actions - for
        # example, Ctrl-Alt-1 might be bound to
        # toggleOverlayList on an OrthoPanel, but bound
        # to toggleTimeSeriesList on a TimeSeriesPanel.
        #
        # The idea is that, when the user presses such a
        # shortcut, the relevant action should be
        # executed on the currently focused ViewPanel
        # (if it has an action associated with the
        # shortcut).
        #
        # The __viewPanelShortcuts dictionary contains
        # mappings from keyboard shortcuts to ViewPanel
        # actions, so we can easily figure out which
        # ViewPanel/action to execute based on the
        # keyboard shortcut.

        viewPanel  = self.focusedViewPanel
        actionName = self.__viewPanelShortcuts[shortcut].get(viewPanel, None)

        if actionName is None:
            return

        func = viewPanel.getAction(actionName)

        if runOnIdle:
            idle.idle(func)
        else:
            func()


    def __onViewPanelClose(self, ev=None, panel=None, displaySync=True):
        """Called when the user closes a :class:`.ViewPanel`. May also
        be called programmatically via the :meth:`removeViewPanel` or
        :removeAllViewPanels` :method.

        The :meth:`__addViewPanelMenu` method adds a *Close* menu item
        for every view panel, and binds it to this method.

        This method does the following:

         1. Makes sure that the ``ViewPanel``: is destroyed correctly
         2. Removes the *Settings* sub-menu corresponding to the ``ViewPanel``.
         3. Makes sure that any remaining ``ViewPanel`` panels are arranged
            nicely.
         4. If the ``displaySync`` parameter is ``True``, calls
            :meth:`__configDisplaySync`.

        :arg ev:          ``wx`` event, passed when a :class:`.ViewPanel` is
                          closed by the user:

        :arg panel:       If called programmatically, the :class:`.ViewPanel`
                          to close.

        :arg displaySync: If ``True`` (the default), and only one ``ViewPanel``
                          remains after this one is removed, that remaining
                          ``ViewPanel`` is synchronised to the master
                          :class:`.DisplayContext`.
        """

        if ev is not None:
            ev.Skip()

            # Undocumented - the window associated with an
            # AuiPaneInfo is available as an attribute called
            # 'window'. Honestly, I don't know why there is
            # not a method available on the AuiPaneInfo or
            # AuiManager to retrieve a managed Window given
            # the associated AuiPaneInfo object.
            paneInfo = ev.GetPane()
            panel    = paneInfo.window

        elif panel is not None:
            paneInfo = self.__auiManager.GetPane(panel)

        # This method may get called when
        # the user closes a control panel
        if panel is None or panel not in self.__viewPanels:
            return

        self              .__viewPanels          .remove(panel)
        self              .__viewPanelIDs        .pop(   panel)
        self              .__viewPanelTitles     .pop(   panel)
        dctx        = self.__viewPanelDCs        .pop(   panel)
        menu        = self.__viewPanelMenus      .pop(   panel, None)
        actionItems = self.__viewPanelMenuActions.pop(   panel, None)

        for action, item in actionItems:
            action.unbindWidget(item)

        for shortcutList in self.__viewPanelShortcuts.values():
            for key in list(shortcutList.keys()):
                if key is panel:
                    shortcutList.pop(key)

        log.debug('Destroying {} ({}) and '
                  'associated DisplayContext ({})'.format(
                      type(panel).__name__,
                      id(panel),
                      id(dctx)))

        # Remove the view panel menu,
        # and make sure that the tools
        # menu is consistent
        if menu is not None:
            self.__settingsMenu.Delete(menu.GetId())

        self.refreshToolsMenu()

        # Calling fslpanel.FSLeyesPanel.destroy()
        # and DisplayContext.destroy() - the
        # AUIManager should do the
        # wx.Window.Destroy side of things ...
        panel.destroy()
        dctx .destroy()

        # If the removed panel was the centre
        # pane, move another panel to the centre
        numPanels = len(self.__viewPanels)
        wasCentre = paneInfo.dock_direction_get() == aui.AUI_DOCK_CENTRE

        if numPanels >= 1 and wasCentre:
            paneInfo = self.__auiManager.GetPane(self.__viewPanels[0])
            paneInfo.Centre()

        # If there is only one panel
        # left, hide its title bar
        if numPanels == 1:
            paneInfo = self.__auiManager.GetPane(self.__viewPanels[0])
            paneInfo.Dockable(False).CaptionVisible(False)

        # if there are no panels,
        # disable the menus
        self.__enableMenus(numPanels > 0)

        if displaySync:
            self.__configDisplaySync()


    def __configDisplaySync(self, newPanel=None):
        """Called by :meth:`addViewPanel` and :meth:`__onViewPanelClose`.

        This method ensures that the display properties, overlay order,
        and selected overlay for the :class:`.DisplayContext` of at least
        least one :class:`.CanvasPanel` is synced to the master
        :class:`.DisplayContext`.

        :arg newPanel: If this method has been called as a result of a new
                       :class:`.ViewPanel` being added, a reference to the
                       new panel.
        """

        import fsleyes.views.canvaspanel as canvaspanel
        import fsleyes.views.plotpanel   as plotpanel

        # Plot panel defaults
        if newPanel is not None and isinstance(newPanel, plotpanel.PlotPanel):
            newPanel.displayCtx.syncOverlayDisplay = True
            newPanel.displayCtx.syncOverlayVolume  = True
            newPanel.displayCtx.unsyncFromParent('overlayOrder')
            newPanel.displayCtx.unsyncFromParent('selectedOverlay')

        canvasPanels = [vp for vp in self.__viewPanels
                        if isinstance(vp, canvaspanel.CanvasPanel)]
        canvasCtxs   = [c.displayCtx for c in canvasPanels]
        numCanvases  = len(canvasPanels)

        # We only care about canvas
        # panels from here on.
        if numCanvases == 0:
            return

        # Is there at least one canvas panel
        # which has display properties/overlay
        # order synced to the master display
        # context?
        displaySynced = any([c.syncOverlayDisplay
                             for c in canvasCtxs
                             if c is not newPanel])
        volumeSynced = any([c.syncOverlayVolume
                             for c in canvasCtxs
                             if c is not newPanel])
        orderSynced   = any([c.isSyncedToParent('overlayOrder')
                             for c in canvasCtxs
                             if c is not newPanel])
        selOvlSynced  = any([c.isSyncedToParent('selectedOverlay')
                             for c in canvasCtxs
                             if c is not newPanel])

        # If an existing canvas panel is
        # already synced to the master,
        # then we set the new panel to
        # be unsynced
        if displaySynced and volumeSynced and orderSynced and selOvlSynced:
            if newPanel is not None and \
               isinstance(newPanel, canvaspanel.CanvasPanel):
                childDC                    = newPanel.displayCtx
                childDC.syncOverlayVolume  = True
                childDC.syncOverlayDisplay = False
                childDC.unsyncFromParent('overlayOrder')
                childDC.unsyncFromParent('selectedOverlay')

        # If no existing CanvasPanels are
        # synced to the master context,
        # re-sync the most recently added
        # one.
        else:
            if newPanel is not None: panel = newPanel
            else:                    panel = canvasPanels[-1]
            childDC = panel.displayCtx

            # Make sure that the parent context
            # inherits the values from this context
            displays = [childDC.getDisplay(o) for o in self.__overlayList]

            childDC.setBindingDirection(False)

            for display in displays:
                opts = display.opts
                display.setBindingDirection(False)
                opts   .setBindingDirection(False)

            childDC.syncOverlayDisplay = True
            childDC.syncOverlayVolume  = True
            childDC.syncToParent('overlayOrder')
            childDC.syncToParent('selectedOverlay')

            # Reset the binding directiona
            childDC.setBindingDirection(True)

            for display in displays:
                opts = display.opts
                display.setBindingDirection(True)
                opts   .setBindingDirection(True)


    def Close(self, **kwargs):
        """Closes this ``FSLeyesFrame``. See :meth:`__onClose`.

        :arg askUnsaved: Defaults to ``True``. If ``False``, the user is not
                         asked whether they want to save any unsaved overlays.

        :arg askLayout:  Defaults to the ``save`` value passed to
                         :meth:`__init__`. Controls whether the user is asked
                         if they want to save the current layout.
        """

        askUnsaved = kwargs.pop('askUnsaved', True)
        askLayout  = kwargs.pop('askLayout',  self.__saveLayout)

        self.__askUnsaved = askUnsaved
        self.__saveLayout = askLayout

        super(FSLeyesFrame, self).Close()


    def __onClose(self, ev):
        """Called when the user closes this ``FSLeyesFrame``.

        Saves the frame position, size, and layout, so it may be preserved the
        next time it is opened. See the :meth:`_restoreState` method.
        """

        import fsleyes.actions.saveoverlay as saveoverlay

        # Check to see if there are any unsaved images
        allSaved = saveoverlay.checkOverlaySaveState(
            self.__overlayList, self.__displayCtx)

        # If there are, get the
        # user to confirm quit
        if not allSaved and self.__askUnsaved:

            msg   = strings.messages[self, 'unsavedOverlays']
            title = strings.titles[  self, 'unsavedOverlays']

            dlg = wx.MessageDialog(self,
                                   message=msg,
                                   caption=title,
                                   style=(wx.YES_NO        |
                                          wx.NO_DEFAULT    |
                                          wx.CENTRE        |
                                          wx.ICON_WARNING))

            dlg.CentreOnParent()
            if dlg.ShowModal() == wx.ID_NO:
                ev.Veto()
                return

        if self.__saveLayout:

            # Ask the user if they want to save the layout,
            # as some people might not like the previous
            # layout being restored on startup.
            save      = fslsettings.read('fsleyes.frame.saveLayout',      True)
            askToSave = fslsettings.read('fsleyes.frame.askToSaveLayout', True)

            if askToSave:

                # Give the user the option of suppressing
                # this dialog forever more
                dlg = fsldlg.CheckBoxMessageDialog(
                    self,
                    strings.titles[self, 'saveLayout'],
                    message=strings.messages[self, 'saveLayout'],
                    cbMessages=[strings.messages[self, 'dontAskToSaveLayout']],
                    cbStates=[False],
                    yesText='Yes',
                    noText='No',
                    cancelText='Cancel',
                    focus='no',
                    icon=wx.ICON_QUESTION)

                result = dlg.ShowModal()

                if result == wx.ID_CANCEL:
                    ev.Veto()
                    return

                save = result == wx.ID_YES

                fslsettings.write('fsleyes.frame.saveLayout', save)
                fslsettings.write('fsleyes.frame.askToSaveLayout',
                                  not dlg.CheckBoxState())

            if save:
                layout   = layouts.serialiseLayout(self)
                size     = self.GetSize().Get()
                position = self.GetScreenPosition().Get()

                log.debug('Saving size: {}'    .format(size))
                log.debug('Saving position: {}'.format(position))
                log.debug('Saving layout: {}'  .format(layout))

                fslsettings.write('fsleyes.frame.size',     size)
                fslsettings.write('fsleyes.frame.position', position)
                fslsettings.write('fsleyes.frame.layout',   layout)
            else:
                fslsettings.delete('fsleyes.frame.size')
                fslsettings.delete('fsleyes.frame.position')
                fslsettings.delete('fsleyes.frame.layout')

        # The close is going ahead. We assume that this
        # EVT_CLOSE handler is the first that is called,
        # and can thus control whether any the event
        # gets propagated to other handlers.
        ev.Skip()

        # It's nice to explicitly clean
        # up our FSLeyesPanels, otherwise
        # they'll probably complain
        for vp in self.__viewPanels:
            self.__onViewPanelClose(panel=vp, displaySync=False)

        self.__auiManager.UnInit()
        self.__auiManager._frame = None
        self.__auiManager        = None

        # (not created) self.__overlayMenuActions

        # Cleanly destroy all
        # menu action objects
        allactions = []
        allactions.extend([a for a, _ in self.__viewMenuActions])
        allactions.extend([a for a, _ in self.__layoutMenuActions])
        allactions.extend([a for a, _ in self.__overlayMenuActions])
        allactions.extend([a for a, _ in self.__toolsMenuActions])
        allactions.extend(self.__menuActions.values())
        for vpactions in self.__viewPanelMenuActions.values():
            allactions.extend([a for a, _ in vpactions])

        for action in allactions:
            # actions associated with some object
            # (e.g. a ViewPanel) will have already
            # been destroyed by them
            if not action.destroyed:
                action.destroy()

        self.__layoutMenuActions    = None
        self.__viewPanelMenuActions = None
        self.__toolsMenuActions     = None
        self.__menuActions          = None

        # Deregister loadoverlay listener
        # (registered in __init__)
        import fsleyes.actions.loadoverlay as loadoverlay
        loadoverlay.recentPathManager.deregister(self.__name)


    def __restoreState(self, restore):
        """Called by :meth:`__init__`.

        If any frame size/layout properties have previously been saved via the
        :mod:`~fsl.utils.settings` module, they are read in, and applied to
        this frame.

        :arg restore: If ``False``, any saved layout state is ignored.
        """

        from operator import itemgetter as iget

        # Restore the saved frame size/position
        size     = fslsettings.read('fsleyes.frame.size')
        position = fslsettings.read('fsleyes.frame.position')
        layout   = fslsettings.read('fsleyes.frame.layout')

        if (size is not None) and (position is not None):

            # Turn the saved size/pos into
            # a (tlx, tly, brx, bry) tuple
            frameRect = [position[0],
                         position[1],
                         position[0] + size[0],
                         position[1] + size[1]]

            # Now make a bounding box containing the
            # space made up of all available displays.
            # Get the bounding rectangles of each
            # display, and change them from
            # (x, y, w, h) into (tlx, tly, brx, bry).
            displays  = [wx.Display(i)   for i in range(wx.Display.GetCount())]
            dispRects = [d.GetGeometry() for d in displays]
            dispRects = [[d.GetTopLeft()[    0],
                          d.GetTopLeft()[    1],
                          d.GetBottomRight()[0],
                          d.GetBottomRight()[1]] for d in dispRects]

            # get the union of these display
            # rectangles (tlx, tly, brx, bry)
            dispRect = [min(dispRects, key=iget(0))[0],
                        min(dispRects, key=iget(1))[1],
                        max(dispRects, key=iget(2))[2],
                        max(dispRects, key=iget(3))[3]]

            # Now we have our two rectangles - the
            # rectangle of our saved frame position,
            # and the rectangle of the available
            # display space.

            # Calculate the area of intersection
            # betwen the two rectangles, and the
            # area of our saved frame position
            xOverlap  = max(0, min(frameRect[2], dispRect[2]) -
                               max(frameRect[0], dispRect[0]))
            yOverlap  = max(0, min(frameRect[3], dispRect[3]) -
                               max(frameRect[1], dispRect[1]))

            intArea   = xOverlap * yOverlap
            frameArea = ((frameRect[2] - frameRect[0]) *
                         (frameRect[3] - frameRect[1]))

            # If the ratio of (frame-display intersection) to
            # (saved frame position) is 'not decent', then
            # forget it, and use a default frame position/size
            ratio = intArea / float(frameArea)
            if ratio < 0.9:

                log.debug('Intersection of saved frame area with available '
                          'display area is too small ({}) - reverting to '
                          'default frame size/position'.format(ratio))

                size     = None
                position = None

        if size is not None:
            log.debug('Restoring previous size: {}'.format(size))
            self.SetSize(size)

        else:

            # Default size is 90% of
            # the first display size
            size    = list(wx.Display(0).GetGeometry().GetSize())
            size[0] = round(size[0] * 0.9)
            size[1] = round(size[1] * 0.9)
            log.debug('Setting default frame size: {}'.format(size))
            self.SetSize(size)

        if position is not None:
            log.debug('Restoring previous position: {}'.format(position))
            self.SetPosition(position)
        else:
            self.Centre()

        if restore:
            if layout is not None:
                log.debug('Restoring previous layout: {}'.format(layout))

                try:
                    layouts.applyLayout(
                        self,
                        'fsleyes.frame.layout',
                        layout,
                        message=strings.messages[self, 'restoringLayout'])
                except Exception:
                    log.warning('Previous layout could not be restored - '
                                'falling back to default layout.')
                    log.debug('Layout restore error', exc_info=True)
                    layout = None

            if layout is None:
                layouts.loadLayout(self, 'default')


    def __makeMenuBar(self):
        """Constructs a bunch of menu items for this ``FSLeyesFrame``."""

        if not self.__haveMenu:
            return

        menuBar = wx.MenuBar()
        self.SetMenuBar(menuBar)

        # The menu bar on OSX/wxPython is a bit different
        # than the menu bar on OSX/wxPhoenix, or on Linux.
        # This is because under OSX/wxPython, we can't get
        # access to the built-in application menu.
        onOSX          = fwidgets.wxPlatform() in (fwidgets.WX_MAC_CARBON,
                                                   fwidgets.WX_MAC_COCOA)
        haveAppMenu    = (onOSX and
                          fwidgets.wxFlavour() == fwidgets.WX_PHOENIX)
        locationOffset = 0

        # On linux, we create a FSLeyes menu
        if not onOSX:
            locationOffset  = 1
            fsleyesMenu     = wx.Menu()
            menuBar.Append(fsleyesMenu, 'FSLeyes')

        # On OSX/wxPhoenix, we can
        # get the built-in app menu
        elif haveAppMenu:
            fsleyesMenu = menuBar.OSXGetAppleMenu()

        # On OSX/wxPython, we fudge
        # things a bit - see below.
        else:
            fsleyesMenu = None

        fileMenu     = wx.Menu()
        overlayMenu  = wx.Menu()
        viewMenu     = wx.Menu()
        settingsMenu = wx.Menu()
        toolsMenu    = wx.Menu()

        menuBar.Append(fileMenu,     'File')
        menuBar.Append(overlayMenu,  'Overlay')
        menuBar.Append(viewMenu,     'View')
        menuBar.Append(settingsMenu, 'Settings')
        menuBar.Append(toolsMenu,    'Tools')

        self.__menuBar      = menuBar
        self.__fsleyesMenu  = fsleyesMenu
        self.__overlayMenu  = overlayMenu
        self.__fileMenu     = fileMenu
        self.__viewMenu     = viewMenu
        self.__layoutMenu   = None
        self.__settingsMenu = settingsMenu
        self.__toolsMenu    = toolsMenu

        # store locations of some menus
        # for use by the __enableMenus
        # method
        self.__menuLocations = {
            'overlay'  : 1 + locationOffset,
            'settings' : 3 + locationOffset,
            'tools'    : 4 + locationOffset,
        }

        self.__makeFileMenu()

        # We have a FSLeyes menu
        if fsleyesMenu is not None:
            self.__makeFSLeyesMenu(fsleyesMenu)

        # We don't have a FSLeyes menu -
        # throw all of the FSLeyes menu
        # stuff onto the end of the File
        # menu.
        else:
            fileMenu.AppendSeparator()
            self.__makeFSLeyesMenu(fileMenu)

        self.__makeOverlayMenu()
        self.refreshToolsMenu()
        self.refreshViewMenu()


    def __enableMenus(self, state):
        """Enables/disables some of the menus. Only the menus which are useless
        without any view panels open are affected.
        """

        # we might be running without a menubar
        # if __init__ was called with menu=Falsee
        if self.__menuBar is None:
            return

        # overlay, settings, and tools menus
        self.__menuBar.EnableTop(self.__menuLocations['overlay'],  state)
        self.__menuBar.EnableTop(self.__menuLocations['settings'], state)
        self.__menuBar.EnableTop(self.__menuLocations['tools'],    state)


    def __makeFSLeyesMenu(self, menu):
        """Called by :meth:`__makeMenuBar`. Creates the *FSLeyes* menu. """

        overlayList = self.__overlayList
        displayCtx  = self.__displayCtx

        from fsleyes.actions.about            import AboutAction
        from fsleyes.actions.diagnosticreport import DiagnosticReportAction
        from fsleyes.actions.clearsettings    import ClearSettingsAction
        from fsleyes.actions.updatecheck      import UpdateCheckAction

        fsleyesActions = [
            (AboutAction(overlayList, displayCtx, self),
             strings.actions[      AboutAction],
             shortcuts.actions.get(AboutAction),
             wx.ID_ABOUT),
            (DiagnosticReportAction(overlayList, displayCtx, self),
             strings.actions[      DiagnosticReportAction],
             shortcuts.actions.get(DiagnosticReportAction),
             wx.ID_ANY),
            (ClearSettingsAction(overlayList, displayCtx, self),
             strings.actions[ClearSettingsAction],
             None,
             wx.ID_ANY),
            (UpdateCheckAction(overlayList, displayCtx),
             strings.actions[      UpdateCheckAction],
             shortcuts.actions.get(UpdateCheckAction),
             wx.ID_ANY),
            (self.setFSLDIR,
             strings.actions[self, 'setFSLDIR'],
             shortcuts.actions.get((self, 'setFSLDIR')),
             wx.ID_ANY),
            (self.openHelp,
             strings.actions[       self, 'openHelp'],
             shortcuts.actions.get((self, 'openHelp')),
             wx.ID_HELP),
            (self.closeFSLeyes,
             strings.actions[       self, 'closeFSLeyes'],
             shortcuts.actions.get((self, 'closeFSLeyes')),
             wx.ID_EXIT)]

        for action, title, shortcut, wxid in fsleyesActions:

            if shortcut is None: shortcut = ''
            else:                shortcut = '\t{}'.format(shortcut)

            title = '{}{}'.format(title, shortcut)

            item = menu.Append(wxid, title)
            action.bindToWidget(self, wx.EVT_MENU, item)

            self.__menuActions[type(action)] = action


    def __makeFileMenu(self):
        """Called by :meth:`__makeMenuBar`. Creates the *File* menu. """

        from fsleyes.actions.loadoverlay        import LoadOverlayAction
        from fsleyes.actions.loadoverlayfromdir import LoadOverlayFromDirAction
        from fsleyes.actions.loaddicom          import LoadDicomAction
        from fsleyes.actions.loadstandard       import LoadStandardAction
        from fsleyes.actions.loadatlas          import LoadAtlasAction
        from fsleyes.actions.browsexnat         import BrowseXNATAction
        from fsleyes.actions.newimage           import NewImageAction
        from fsleyes.actions.runscript          import RunScriptAction
        from fsleyes.actions.loadplugin         import LoadPluginAction
        from fsleyes.actions.notebook           import NotebookAction

        menu = self.__fileMenu

        fileActions = [LoadOverlayAction,
                       LoadOverlayFromDirAction,
                       LoadStandardAction,
                       NewImageAction,
                       LoadDicomAction,
                       BrowseXNATAction,
                       LoadAtlasAction,
                       NotebookAction,
                       LoadPluginAction,
                       RunScriptAction]

        for action in fileActions:

            if action == 'sep':
                menu.AppendSeparator()
                continue

            title    = strings.actions[  action]
            shortcut = shortcuts.actions.get(action)

            if shortcut is not None:
                title = '{}\t{}'.format(title, shortcut)

            menuItem  = menu.Append(wx.ID_ANY, title)
            actionObj = action(self.__overlayList, self.__displayCtx, self)

            self.__menuActions[action] = actionObj

            actionObj.bindToWidget(self, wx.EVT_MENU, menuItem)

        # The last option in the file
        # menu is a 'Recent files' list.
        # The list gets populated when
        # it is clicked.
        self.__recentPathsMenu = wx.Menu()
        menu.AppendSubMenu(self.__recentPathsMenu,
                           strings.labels[self, 'recentPathsMenu'])
        self.__makeRecentPathsMenu()


    def __makeRecentPathsMenu(self, *a):
        """Populates the "File -> Recent files" menu. This method is called
        by the :meth:`__makeFileMenu`, and also by the
        :class:`.RecentPathManager` when paths are added.
        """

        if not self.__haveMenu:
            return

        import fsleyes.actions.loadoverlay as loadoverlay

        for path in self.__recentPathsMenu.GetMenuItems():
            self.__recentPathsMenu.Remove(path)

        paths = loadoverlay.recentPathManager.listRecentPaths()

        # We give each recent file menu item an identifying
        # number from 1-(numpaths) - this is used by the
        # __onRecentPath handler.
        for i, p in enumerate(reversed(paths), start=1):
            menuItem = self.__recentPathsMenu.Append(i, p)
            self.Bind(wx.EVT_MENU, self.__onRecentPath, menuItem)


    def __onRecentPath(self, ev):
        """Called when a recent path is selected from the
        "File -> Recent files" menu item. Loads the path.
        See the :meth:`__makeRecentPathsMenu` method.
        """

        import fsleyes.actions.loadoverlay as loadoverlay

        path = self.__recentPathsMenu.GetLabel(ev.GetId())

        def onLoad(paths, overlays):

            if len(overlays) == 0:
                return

            self.__overlayList.extend(overlays)

            self.__displayCtx.selectedOverlay = \
                self.__displayCtx.overlayOrder[-1]

            if self.__displayCtx.autoDisplay:
                for overlay in overlays:
                    autodisplay.autoDisplay(overlay,
                                            self.__overlayList,
                                            self.__displayCtx)

        loadoverlay.loadOverlays([path],
                                 onLoad=onLoad,
                                 inmem=self.__displayCtx.loadInMemory)


    def __makeViewMenu(self):
        """Called by :meth:`refreshViewMenu`. Creates the view panel menu.

        :returns: A list containing the ``(Action, wx.MenuItem)`` pairs that
                  were added to the menu.
        """

        menu = self.__viewMenu

        # Shortcuts to open a new view panel
        vpActions = [self.addOrthoPanel,
                     self.addLightBoxPanel,
                     self.addTimeSeriesPanel,
                     self.addHistogramPanel,
                     self.addPowerSpectrumPanel,
                     self.addScene3DPanel,
                     self.addShellPanel]

        actionItems = []

        for action in vpActions:

            shortcut = shortcuts.actions.get((self, action.__name__))

            if shortcut is None: shortcut = ''
            else:                shortcut = '\t{}'.format(shortcut)

            title = '{}{}'.format(strings.actions[self, action.__name__],
                                  shortcut)

            item = menu.Append(wx.ID_ANY, title)
            action.bindToWidget(self, wx.EVT_MENU, item)
            actionItems.append((action, item))

        # Views provided by FSLeyes plugins
        pluginViews = plugins.listViews()
        if len(pluginViews) > 0:
            menu.AppendSeparator()
            for name, cls in pluginViews.items():
                func = ft.partial(self.addViewPanel, cls, title=name)
                item = menu.Append(wx.ID_ANY, name)
                self.Bind(wx.EVT_MENU, lambda ev, f=func : f(), item)

        # We create the layout menu,
        # but don't populate it - this
        # is performed  by the caller
        self.__layoutMenu = wx.Menu()
        menu.AppendSeparator()
        menu.AppendSubMenu(self.__layoutMenu, 'Layouts')
        return actionItems


    def __makeLayoutMenu(self):
        """Called by :meth:`refreshLayoutMenu`. Re-creates the *View->Layouts*
        menu.
        """

        if not self.__haveMenu:
            return []

        from fsleyes.actions.loadlayout   import LoadLayoutAction
        from fsleyes.actions.savelayout   import SaveLayoutAction
        from fsleyes.actions.clearlayouts import ClearLayoutsAction

        menu        = self.__layoutMenu
        actionItems = []

        builtIns = list(layouts.BUILT_IN_LAYOUTS.keys())
        saved    = layouts.getAllLayouts()

        # Add a menu item to load each built-in layouts
        for layout in builtIns:

            title    = strings.layouts.get(layout, layout)
            shortcut = shortcuts.actions.get((self, 'layouts', layout),
                                             None)

            if shortcut is not None:
                title = '{}\t{}'.format(title, shortcut)

            menuItem = menu.Append(wx.ID_ANY, title)

            actionObj = LoadLayoutAction(
                self.overlayList, self.displayCtx, self, layout)
            actionObj.bindToWidget(self, wx.EVT_MENU, menuItem)
            actionItems.append((actionObj, menuItem))

        if len(builtIns) > 0:
            menu.AppendSeparator()

        # Add a menu item to load each saved layout
        for layout in saved:

            menuItem  = menu.Append(
                wx.ID_ANY, strings.layouts.get(layout, layout))
            actionObj = LoadLayoutAction(
                self.overlayList, self.displayCtx, self, layout)
            actionObj.bindToWidget(self, wx.EVT_MENU, menuItem)

            actionItems.append((actionObj, menuItem))

        # Add menu items for other layout
        # operations, but separate them
        # from the existing layouts
        if len(saved) > 0:
            menu.AppendSeparator()

        # TODO: Delete a single layout?
        #       Save to/load from file?
        layoutActions = [SaveLayoutAction,
                         ClearLayoutsAction]

        for la in layoutActions:

            actionObj = la(self.overlayList, self.displayCtx, self)
            menuItem  = menu.Append(wx.ID_ANY, strings.actions[la])

            actionObj.bindToWidget(self, wx.EVT_MENU, menuItem)
            actionItems.append((actionObj, menuItem))
        return actionItems


    def __makeOverlayMenu(self):
        """Called by :meth:`__makeMenuBar`. Creates/configures the *Overlay*
        menu.
        """

        from fsleyes.actions.removealloverlays import RemoveAllOverlaysAction
        from fsleyes.actions.copyoverlay       import CopyOverlayAction
        from fsleyes.actions.saveoverlay       import SaveOverlayAction
        from fsleyes.actions.reloadoverlay     import ReloadOverlayAction
        from fsleyes.actions.removeoverlay     import RemoveOverlayAction

        menu = self.__overlayMenu

        fileActions = ['selectNextOverlay',
                       'selectPreviousOverlay',
                       RemoveAllOverlaysAction,
                       'sep',
                       'name',
                       CopyOverlayAction,
                       SaveOverlayAction,
                       ReloadOverlayAction,
                       RemoveOverlayAction,
                       'toggleOverlayVisibility']

        for action in fileActions:

            # A dummy menu item which contains the
            # name of the currently selected overlay.
            if action == 'name':
                self.__overlayNameMenuItem = menu.Append(wx.ID_ANY, '<')
                self.__overlayNameMenuItem.Enable(False)
                continue

            if action == 'sep':

                menu.AppendSeparator()
                continue

            # Method on this FSLeyesFrame
            if isinstance(action, str):

                title     = strings.actions[  self, action]
                shortcut  = shortcuts.actions.get((self, action))
                actionObj = getattr(self, action)

            # Action class
            else:
                title     = strings.actions[  action]
                shortcut  = shortcuts.actions.get(action)
                actionObj = action(self.__overlayList, self.__displayCtx, self)

            if shortcut is not None:
                title = '{}\t{}'.format(title, shortcut)

            menuItem = menu.Append(wx.ID_ANY, title)
            actionObj.bindToWidget(self, wx.EVT_MENU, menuItem)
            self.__overlayMenuActions.append((actionObj, menuItem))


    def __makeToolsMenu(self):
        """Called by :meth:`refreshToolsMenu`. Populates the *Tools* menu with
        tools that are not bound to a specific ``ViewPanel`` (and which are
        always present), followed by tools which are specific to the open
        views.
        """

        menu        = self.__toolsMenu
        actionItems = []
        pluginTools = plugins.listTools()

        # Tools are either restricted for use with a
        # specific view, or independent of any view.
        # We create the latter here first, and then
        # use __makeViewPanelTools to create the former
        pluginTools = {n : c for n, c in pluginTools.items()
                       if c.supportedViews() is None}

        # Fix the ordering for view-independent plugins
        # (the equivalent of what is done with
        # ViewPanel.toolOrder, in __makeViewPanelTools).
        toolOrder   = ['ApplyFlirtXfmAction',
                       'SaveFlirtXfmAction',
                       'ResampleAction',
                       'ProjectImageToSurfaceAction']
        names, clss = zip(*pluginTools.items())
        toolOrder   = [plugins.lookupTool(t) for t in toolOrder]
        indices     = [toolOrder.index(c) if c in toolOrder
                       else len(pluginTools) for c in clss]
        pluginTools = sorted(zip(indices, names, clss))
        pluginTools = {t[1]: t[2] for t in pluginTools}

        for name, cls in pluginTools.items():

            # Refs to view-independent tools are stored in
            # __menuActions. This method may be called
            # multiple times, so we only create the Action
            # object on the first invocation.
            if cls in self.__menuActions:
                actionObj = self.__menuActions[cls]
            else:
                # plugin tools not coupled to a specific view
                # get a reference to me, the FSLeyesFrame
                actionObj = cls(self.__overlayList, self.__displayCtx, self)

            shortcut = shortcuts.actions.get(cls)

            if shortcut is not None:
                name = '{}\t{}'.format(name, shortcut)

            menuItem = menu.Append(wx.ID_ANY, name)
            actionObj.bindToWidget(self, wx.EVT_MENU, menuItem)

            actionItems.append((actionObj, menuItem))

        actionItems.extend(self.__makeViewPanelTools())
        return actionItems


    def __makeViewPanelTools(self):
        """Called when a view panel is added or removed. Refreshes the *Tools*
        menu.
        """

        if not self.__haveMenu:
            return

        menu = self.__toolsMenu

        from fsleyes.views.orthopanel         import OrthoPanel
        from fsleyes.views.lightboxpanel      import LightBoxPanel
        from fsleyes.views.scene3dpanel       import Scene3DPanel
        from fsleyes.views.timeseriespanel    import TimeSeriesPanel
        from fsleyes.views.histogrampanel     import HistogramPanel
        from fsleyes.views.powerspectrumpanel import PowerSpectrumPanel
        from fsleyes.views.shellpanel         import ShellPanel

        # Recreate tools for each view panel. We
        # ensure that the tools for different view
        # panel types are always in a consistent
        # order (i.e. ortho first, shell last)
        panels  = self.viewPanels
        vpOrder = [OrthoPanel,
                   LightBoxPanel,
                   Scene3DPanel,
                   TimeSeriesPanel,
                   HistogramPanel,
                   PowerSpectrumPanel,
                   ShellPanel]
        for p in panels:
            if type(p) not in vpOrder:
                vpOrder.append(type(p))
        panels = sorted(panels, key=lambda p: vpOrder.index(type(p)))

        actionItems  = []
        vpTypesAdded = set()

        for panel in panels:

            vpType = type(panel)

            # built-in tools, implemented
            # by the viewpanel itself
            toolNames  = [t.actionName for t in panel.getTools()]
            toolTitles = {}

            # Plugin-provided tools, which are created
            # by the ViewPanel and added as attributes
            # to itself (see ViewPanel.reloadPlugins)
            pluginTools = plugins.listTools(vpType)

            # ViewPanel.toolOrder can suggest an
            # ordering of plugin-provided tools.
            toolOrder = vpType.toolOrder()
            if len(pluginTools) > 0 and toolOrder is not None:
                names, clss = zip(*pluginTools.items())
                toolOrder   = [plugins.lookupTool(t) for t in toolOrder]
                indices     = [toolOrder.index(c) if c in toolOrder
                               else len(pluginTools) for c in clss]
                pluginTools = sorted(zip(indices, names, clss))
                pluginTools = {t[1] : t[2] for t in pluginTools}

            # See ViewPanel.reloadPlugins. and LoadPluginAction.
            # All supported tools are added as attributes to the
            # ViewPanel instance, with the class name used as
            # the attribute name.
            for toolName, cls in pluginTools.items():
                name = cls.__name__
                toolNames.append(name)
                toolTitles[name] = toolName

            # Only the first panel for each type
            # has its tools added to the menu.
            if len(toolNames) == 0 or vpType in vpTypesAdded:
                continue

            vpTypesAdded.add(vpType)

            # Each view panel added to the tools list
            # gets its own section, starting with a
            # separator, and the view panel title.
            menu.AppendSeparator()
            menu.Append(wx.ID_ANY, self.__viewPanelTitles[panel]).Enable(False)
            actionItems.extend(
                self.populateMenu(menu, panel, toolNames, toolTitles))

        return actionItems


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay`, or
        when the :attr:`.Display.name` property of the currently selected
        overlay changes. Updates the overlay name item in the *Overlay*
        menu*.
        """

        oldDisplay = getattr(self, '__display', None)

        if oldDisplay is not None:
            oldDisplay.removeListener('name', self.__name)
            self.__display = None

        overlay = self.__displayCtx.getSelectedOverlay()

        haveOverlays = overlay is not None

        self.selectNextOverlay      .enabled = haveOverlays
        self.selectPreviousOverlay  .enabled = haveOverlays
        self.toggleOverlayVisibility.enabled = haveOverlays

        if not self.__haveMenu:
            return

        if haveOverlays:

            self.__display = self.__displayCtx.getDisplay(overlay)
            name = self.__display.name

            if name is None: name = ''
            else:            name = name.strip()

            if name == '':
                name = strings.labels[self, 'noName']
            self.__overlayNameMenuItem.SetItemLabel(name)

            if not self.__display.hasListener('name', self.__name):
                self.__display.addListener('name',
                                           self.__name,
                                           self.__selectedOverlayChanged)
        else:
            name = strings.labels[self, 'noOverlays']
            self.__overlayNameMenuItem.SetItemLabel(name)


class OverlayDropTarget(wx.FileDropTarget):
    """The ``OverlayDropTarget`` class allows overly files to be dragged and
    dropped onto a ``wx`` window. It uses the :func:`.loadOverlays` function.

    Associate an ``OverlayDropTarget`` `dt`` with a window ``w`` like so::

        w.SetDropTarget(dt)
    """


    def __init__(self, overlayList, displayCtx):
        """Create an ``OverlayDropTarget``. The master
        :class:`.DissplayContext` should be passed in, as opposed to any child
        instances.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  Thw master :class:`.DisplayConext`.
        """
        wx.FileDropTarget.__init__(self)
        self.__overlayList = overlayList
        self.__displayCtx  = displayCtx


    def __onLoad(self, paths, overlays):
        """Called by :func:`.loadOverlays` (which is called in
        :meth:`OnDropFiles`) when the dropped overlay files have been loaded.
        Adds the overlays to the :class:`.OverlayList`.
        """

        if len(overlays) == 0:
            return

        self.__overlayList.extend(overlays)

        if self.__displayCtx.autoDisplay:
            for overlay in overlays:
                autodisplay.autoDisplay(overlay,
                                        self.__overlayList,
                                        self.__displayCtx)


    def OnDropFiles(self, x, y, filenames):
        """Overrides ``wx.FileDropTarget.OnDropFiles``. Called when files
        are dropped onto the window. Passes the files to :func:`.loadOverlays`,
        along with :meth:`__onLoad` as a callback.
        """
        import fsleyes.actions.loadoverlay as loadoverlay

        if filenames is not None:
            loadoverlay.loadOverlays(
                filenames,
                onLoad=self.__onLoad,
                inmem=self.__displayCtx.loadInMemory)
            return True
        else:
            return False



#!/usr/bin/env python
#
# offscreenlightboxcanvas.py - The OffScreenLightBoxCanvas.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OffScreenLightBoxCanvas`, which is a
:class:`.LightBoxCanvas` for use in off-screen OpenGL rendering.
"""


import fsleyes.gl as fslgl
from . import        lightboxcanvas


class OffScreenLightBoxCanvas(lightboxcanvas.LightBoxCanvas,
                              fslgl.OffScreenCanvasTarget):
    """The ``OffScreenLightBoxCanvas`` is a :class:`.LightBoxCanvas` which uses
    a :class:`.RenderTexture` as its target, for static off-screen Open GL
    rendering.
    """

    def __init__(self,
                 overlayList,
                 displayCtx,
                 zax,
                 *args,
                 **kwargs):
        """Create an ``OffScreenLightBoxCanvas``.

        See the :class:`.SliceCanvas` and :class:`.OffscreenCanvasTarget`
        classes for details on the other arguments.
        """

        fslgl.OffScreenCanvasTarget  .__init__(self, *args, **kwargs)
        lightboxcanvas.LightBoxCanvas.__init__(self,
                                               overlayList,
                                               displayCtx,
                                               zax)

        # LightBoxCanvas only updates itself
        # when a slice property changes. So
        # we force an initialisation just
        # in case.
        self._slicePropsChanged()



#!/usr/bin/env python
#
# lookuptablepanel.py - The LookupTablePanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LookupTablePanel`, a *FSLeyes control*
panel which allows the user to manage lookup tables. See the
:mod:`.colourmaps` module for more details on lookup tables.

A few other classes and functions are defined in this module, all for use by
the ``LookupTablePanel``:

.. autosummary::
   :nosignatures:

   promptForLutName
   LabelWidget
   LutLabelDialog
"""


import            os
import os.path as op
import            logging

import wx

import fsl.utils.idle                as idle
import fsl.utils.settings            as fslsettings
import fsl.data.melodicimage         as fslmelimage

import fsleyes_props                 as props
import fsleyes_widgets.elistbox      as elistbox
import fsleyes_widgets.utils.status  as status

import fsleyes.views.canvaspanel     as canvaspanel
import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.displaycontext        as displayctx
import fsleyes.colourmaps            as fslcmaps
import fsleyes.strings               as strings



log = logging.getLogger(__name__)


class LookupTablePanel(ctrlpanel.ControlPanel):
    """A ``LookupTablePanel`` is a :class:`.ControlPanel` which allows users
    to manage ``LookupTable`` instances. A ``LookupTablePanel`` looks
    something like this:

    .. image:: images/lookuptablepanel.png
       :scale: 50%
       :align: center


    A ``LookupTablePanel`` allows the user to do the following:

      - Add/remove labels to/from a :class:`LookupTable`.

      - Change the colour, name, and visibility of a label in a
        ``LookupTable``.

      - Create a new ``LookupTable``, or copy an existing one.

      - Save/load a ``LookupTable`` to/from a file.


    The ``LookupTablePanel`` keeps track of the currently selected overlay
    (see the :attr:`.DisplayContext.selectedOverlay` property). If the overlay
    is associated with a :class:`.LabelOpts` instance, its
    :attr:`.Labelopts.lut` property will be updated when the
    :class:`LookupTable` is changed through the ``LookupTablePanel``, and vice
    versa.
    """


    @staticmethod
    def supportedViews():
        """The :class:`.LookupTablePanel` is restricted for use with
        ``OrthoPanel``, ``LightBoxPanel`` and ``Scene3DPanel`` views.
        """
        return [canvaspanel.CanvasPanel]


    @staticmethod
    def defaultLayout():
        """Returns a dictionary containing layout settings to be passed to
        :class:`.ViewPanel.togglePanel`.
        """
        return {'location' : wx.RIGHT}


    def __init__(self, parent, overlayList, displayCtx, viewPanel):
        """Create a ``LookupTablePanel``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg viewPanel:   The :class:`.ViewPanel` instance.
        """

        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, viewPanel)

        self.__controlCol = wx.Panel(self)
        self.__labelList  = elistbox.EditableListBox(
            self, style=(elistbox.ELB_NO_MOVE   |
                         elistbox.ELB_NO_ADD    |
                         elistbox.ELB_NO_REMOVE |
                         elistbox.ELB_WIDGET_RIGHT))

        self.__lutChoice       = wx.Choice(self.__controlCol)
        self.__selAllButton    = wx.Button(self.__controlCol)
        self.__selNoneButton   = wx.Button(self.__controlCol)
        self.__addLabelButton  = wx.Button(self.__controlCol)
        self.__rmLabelButton   = wx.Button(self.__controlCol)
        self.__newLutButton    = wx.Button(self.__controlCol)
        self.__copyLutButton   = wx.Button(self.__controlCol)
        self.__saveLutButton   = wx.Button(self.__controlCol)
        self.__loadLutButton   = wx.Button(self.__controlCol)

        self.__controlColSizer = wx.BoxSizer(wx.VERTICAL)
        self.__sizer           = wx.BoxSizer(wx.VERTICAL)

        self.__controlCol.SetSizer(self.__controlColSizer)
        self             .SetSizer(self.__sizer)

        self.__controlColSizer.Add(self.__lutChoice,      flag=wx.EXPAND)
        self.__controlColSizer.Add(self.__selAllButton,   flag=wx.EXPAND)
        self.__controlColSizer.Add(self.__selNoneButton,  flag=wx.EXPAND)
        self.__controlColSizer.Add(self.__addLabelButton, flag=wx.EXPAND)
        self.__controlColSizer.Add(self.__rmLabelButton,  flag=wx.EXPAND)
        self.__controlColSizer.Add(self.__newLutButton,   flag=wx.EXPAND)
        self.__controlColSizer.Add(self.__copyLutButton,  flag=wx.EXPAND)
        self.__controlColSizer.Add(self.__loadLutButton,  flag=wx.EXPAND)
        self.__controlColSizer.Add(self.__saveLutButton,  flag=wx.EXPAND)

        self.__sizer.Add(self.__controlCol, flag=wx.EXPAND)
        self.__sizer.Add(self.__labelList,  flag=wx.EXPAND, proportion=1)

        # Label the buttons
        self.__selAllButton  .SetLabel(strings.labels[  self, 'selectAll'])
        self.__selNoneButton .SetLabel(strings.labels[  self, 'selectNone'])
        self.__addLabelButton.SetLabel(strings.labels[  self, 'addLabel'])
        self.__rmLabelButton .SetLabel(strings.labels[  self, 'removeLabel'])
        self.__newLutButton  .SetLabel(strings.labels[  self, 'newLut'])
        self.__copyLutButton .SetLabel(strings.labels[  self, 'copyLut'])
        self.__loadLutButton .SetLabel(strings.labels[  self, 'loadLut'])
        self.__saveLutButton .SetLabel(strings.labels[  self, 'saveLut'])

        self.__lutChoice     .Bind(wx.EVT_CHOICE, self.__onLutChoice)
        self.__selAllButton  .Bind(wx.EVT_BUTTON, self.__onSelectAll)
        self.__selNoneButton .Bind(wx.EVT_BUTTON, self.__onSelectNone)
        self.__addLabelButton.Bind(wx.EVT_BUTTON, self.__onLabelAdd)
        self.__rmLabelButton .Bind(wx.EVT_BUTTON, self.__onLabelRemove)
        self.__newLutButton  .Bind(wx.EVT_BUTTON, self.__onNewLut)
        self.__copyLutButton .Bind(wx.EVT_BUTTON, self.__onCopyLut)
        self.__loadLutButton .Bind(wx.EVT_BUTTON, self.__onLoadLut)
        self.__saveLutButton .Bind(wx.EVT_BUTTON, self.__onSaveLut)

        # The selected overlay / DisplayOpts
        # are tracked if they use an LUT
        # (e.g. LabelOpts). And the currently
        # selected/displayed LUT is always
        # tracked.
        self.__selectedOverlay = None
        self.__selectedOpts    = None
        self.__selectedLut     = None

        # See the __createLabelList method
        self.__labelListCreateKey = 0

        overlayList.addListener('overlays',
                                self.name,
                                self.__selectedOverlayChanged)
        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__selectedOverlayChanged)

        self.__updateLutChoices()
        self.__selectedOverlayChanged()

        # If the selected lut was not set
        # via the selectedOverlayChanged
        # call, we'll manually set it here
        if self.__selectedLut is None:
            self.__setLut(fslcmaps.getLookupTables()[0])

        self.Layout()

        # The label values (used as labels in the
        # EditableListBox) seem to get squashed
        # easily. So we're adding a bit of padding
        # to force this panel to have a minimum
        # size, to ensure that the labels are
        # visible.
        dc      = wx.ClientDC(self.__labelList)
        w, h    = dc.GetTextExtent('9999')
        minSize = self.__sizer.GetMinSize()
        self.SetMinSize((minSize[0] + w, minSize[1]))


    def destroy(self):
        """Must be called when this ``LookupTablePanel`` is no longer needed.
        Removes some property listeners, and calls the
        :meth:`ControlPanel.destroy` method.
        """

        self.overlayList.removeListener('overlays',        self.name)
        self.displayCtx .removeListener('selectedOverlay', self.name)

        overlay = self.__selectedOverlay
        opts    = self.__selectedOpts
        lut     = self.__selectedLut

        if overlay is not None and overlay in self.overlayList:

            display = self.displayCtx.getDisplay(overlay)
            display.removeListener('overlayType', self.name)

        if opts is not None:
            opts.removeListener('lut', self.name)

        if lut is not None:
            lut.deregister(self.name, 'saved')
            lut.deregister(self.name, 'added')
            lut.deregister(self.name, 'removed')

        self.__selectedOverlay = None
        self.__selectedOpts    = None
        self.__selectedLut     = None

        ctrlpanel.ControlPanel.destroy(self)


    def __updateLutChoices(self):
        """Refreshes the contents of the lookup table drop down box,
        using the :class:`.LookupTable` instances returned by the
        :func:`.colourmaps.getLookupTables` function.
        """

        log.debug('Updating lookup table choices')

        oldNames     = self.__lutChoice.GetItems()
        oldSelection = self.__lutChoice.GetSelection()

        luts     = fslcmaps.getLookupTables()
        newNames = [l.name for l in luts]

        try:              newSelection = oldNames.index(oldNames[oldSelection])
        except Exception: newSelection = 0

        self.__lutChoice.SetItems(newNames)
        self.__lutChoice.SetSelection(newSelection)

        for i, lut in enumerate(luts):
            self.__lutChoice.SetClientData(i, lut)


    def __createLabelList(self):
        """Refreshes the contents of the class:`.LookupTable` label list, from
        the currently selected ``LookupTable``.
        """

        # The label list is created asynchronously on
        # the wx.Idle loop, because it can take some
        # time for big lookup tables. In the event
        # that the list needs to be re-created (e.g.
        # the current lookup table is changed), this
        # attribute is used so that scheduled creation
        # routines (the addLabel function defined
        # below) can tell whether they should cancel.
        myCreateKey = (self.__labelListCreateKey + 1) % 65536
        self.__labelListCreateKey = myCreateKey

        lut     = self.__selectedLut
        nlabels = len(lut)

        self.__labelList.Clear()

        # If this is a new lut, it
        # won't have any labels
        if nlabels == 0:
            return

        # The label widgets are created via consecutive
        # calls to addLabel, which is scheduled on the
        # idle.idle loop. We create blockSize labels
        # in each asynchronous call.
        blockSize = 100

        def addLabel(startIdx):

            # If the user closes this panel while the
            # label list is being created, wx will
            # complain when we try to append things
            # to a widget that has been destroyed.
            try:

                for labelIdx in range(
                        startIdx, min(startIdx + blockSize, nlabels)):

                    # A new request to re-create the list has
                    # been made - cancel this creation chain.
                    if self.__labelListCreateKey != myCreateKey:
                        return

                    label  = lut[labelIdx]
                    widget = LabelWidget(self, lut, label)

                    self.__labelList.Append(str(label.value),
                                            clientData=label.value,
                                            extraWidget=widget)

                if labelIdx == nlabels - 1:
                    status.update('Lookup table label list created.')
                    self.Enable()
                    self.__labelList.Enable()
                else:
                    idle.idle(addLabel, labelIdx + 1)

            except RuntimeError:
                pass

        log   .debug( 'Creating lookup table label list')
        status.update('Creating lookup table label list...', timeout=None)

        self.__labelList.Disable()
        self.Disable()
        idle.idle(addLabel, 0)


    def __setLut(self, lut):
        """Updates this ``LookupTablePanel`` to display the labels for the
        given ``lut`` (assumed to be a :class:`.LookupTable` instance).

        If the currently selected overlay is associated with a
        :class:`.LabelOpts` instance, its :attr:`.LabelOpts.lut` property is
        set to the new ``LookupTable``.
        """

        if self.__selectedLut == lut:
            return

        log.debug('Selecting lut: {}'.format(lut))

        if self.__selectedLut is not None:
            self.__selectedLut.deregister(self.name, 'saved')
            self.__selectedLut.deregister(self.name, 'added')
            self.__selectedLut.deregister(self.name, 'removed')

        self.__selectedLut = lut

        if lut is not None:
            lut.register(self.name, self.__lutSaveStateChanged, 'saved')
            lut.register(self.name, self.__lutLabelAdded,       'added')
            lut.register(self.name, self.__lutLabelRemoved,     'removed')

        if lut is not None and self.__selectedOpts is not None:
            with props.skip(self.__selectedOpts, 'lut', self.name):
                self.__selectedOpts.lut = lut

        allLuts = fslcmaps.getLookupTables()

        self.__lutChoice.SetSelection(allLuts.index(lut))

        self.__lutSaveStateChanged()
        self.__createLabelList()


    def __lutSaveStateChanged(self, *a):
        """Called when the :attr:`.LookupTable.saved` property of the
        current :class:`LookupTable` instance changes. Sets the state
        of the *save* button accordingly.
        """
        self.__saveLutButton.Enable(not self.__selectedLut.saved)


    def __lutLabelAdded(self, lut, topic, label):
        """Called when the current :class:`.LookupTable` sends an ``'added'``
        notification, indicating that a label has been added. Updates the list
        of displayed labels.
        """
        label, idx = label
        widget     = LabelWidget(self, lut, label)
        self.__labelList.Insert(str(label.value),
                                idx,
                                clientData=label.value,
                                extraWidget=widget)


    def __lutLabelRemoved(self, lut, topic, label):
        """Called when the current :class:`.LookupTable` sends a ``'removed'``
        notification, indicating that a label has been removed. Updates the
        list of displayed labels.
        """
        label, idx = label
        self.__labelList.Delete(idx)


    def __onLutChoice(self, ev):
        """Called when the user changes the selected :class:`.LookupTable`
        via the lookup table drop down box. See the :meth:`__setLut` method..
        """

        selection = self.__lutChoice.GetSelection()
        lut       = self.__lutChoice.GetClientData(selection)

        log.debug('Lut choice: {}'.format(lut))

        self.__setLut(lut)


    def __onSelectAll(self, ev):
        """Called when the user pushes the *Select all* button. Enables
        every label on the current LUT.
        """

        for label in self.__selectedLut:
            label.enabled = True


    def __onSelectNone(self, ev):
        """Called when the user pushes the *Select none* button. Disables
        every label on the current LUT.
        """

        for label in self.__selectedLut:
            label.enabled = False


    def __onNewLut(self, ev):
        """Called when the user presses the *New LUT* button.

        Prompts the user to enter a name (via :func:`promptForLutName`), and
        then creates and registers a new :class:`.LookupTable`
        instance. Updates this ``LookupTablePanel`` via the
        :meth:`__updateLutChoices` and :meth:`__setLut` methods.
        """

        key, name = promptForLutName()

        if key is None:
            return

        log.debug('Creating and registering new '
                  'LookupTable: {}'.format(key))

        lut = fslcmaps.LookupTable(key=key, name=name)
        fslcmaps.registerLookupTable(lut, self.overlayList, self.displayCtx)

        self.__updateLutChoices()
        self.__setLut(lut)


    def __onCopyLut(self, ev):
        """Called when the user presses the *Copy LUT* button.

        Prompts the user to enter a name (via :func:`promptForLutName`), and
        then creates and registers a new :class:`.LookupTable` instance which
        is initialised with the same labels as the previously selected
        ``LookupTable``. Updates this ``LookupTablePanel`` via the
        :meth:`__updateLutChoices` and :meth:`__setLut` methods.
        """

        oldKey  = self.__selectedLut.key
        oldName = self.__selectedLut.name

        newKey, newName = promptForLutName('{} copy'.format(oldName))

        if newKey is None:
            return

        log.debug('Creating and registering new '
                  'LookupTable {} (copied from {})'.format(newKey, oldKey))

        lut = fslcmaps.LookupTable(key=newKey, name=newName)

        for label in self.__selectedLut.labels():
            lut.insert(label.value,
                       name=label.name,
                       colour=label.colour,
                       enabled=label.enabled)

        fslcmaps.registerLookupTable(lut, self.overlayList, self.displayCtx)

        self.__updateLutChoices()
        self.__setLut(lut)


    def __onLoadLut(self, ev):
        """Called when the user presses the *Load LUT* button.  Does the
        following:

          - Prompts the user to select a LUT file with a ``wx.FileDialog``

          - Prompts the user to enter a name for the LUT via the
            :func:`promptForLutName` function.

          - Creates and registers a new :class:`.LookupTable` instance,
            initialising it with the selected file.

          - Updates this ``LookupTablePanel`` via the
            :meth:`__updateLutChoices` and :meth:`__setLut` methods.
        """

        parent  = wx.GetApp().GetTopWindow()
        loadDir = fslsettings.read('fsleyes.loadlutdir', os.getcwd())

        # Prompt the user to select a lut file
        fileDlg = wx.FileDialog(
            parent,
            defaultDir=loadDir,
            message=strings.titles[self, 'loadLut'],
            style=wx.FD_OPEN)

        if fileDlg.ShowModal() != wx.ID_OK:
            return

        lutFile = fileDlg.GetPath()
        lutDir  = op.dirname(lutFile)
        lutName = op.splitext(op.basename(lutFile))[0]

        # Prompt the user to enter a name
        lutKey, lutName = promptForLutName(lutName)
        if lutKey is None:
            return

        # Register the lut
        errTitle = strings.titles[  self, 'loadError']
        errMsg   = strings.messages[self, 'loadError'].format(lutFile)
        with status.reportIfError(errTitle, errMsg):
            lut = fslcmaps.registerLookupTable(lutFile,
                                               self.overlayList,
                                               self.displayCtx,
                                               key=lutKey,
                                               name=lutName)

        # Save the directory for next time
        fslsettings.write('fsleyes.loadlutdir', lutDir)

        # Select the lut in the panel
        self.__updateLutChoices()
        self.__setLut(lut)


    def __onSaveLut(self, ev):
        """Called when the user presses the *Save LUT* button. Makes sure
        that the current :class:`LookupTable` is saved (see the
        :func:`.colourmaps.installLookupTable` function).
        """
        etitle = strings.titles[  self, 'installerror']
        emsg   = strings.messages[self, 'installerror']
        with status.reportIfError(title=etitle, msg=emsg, raiseError=False):
            fslcmaps.installLookupTable(self.__selectedLut.key)


    def __onLabelAdd(self, ev):
        """Called when the user pushes the *add* button on the lookup table
        label list. Displays a :class:`LutLabelDialog`, prompting the user
        to select a name, value and colour, and then adds a new label to the
        current :class:`.LookupTable` instance.
        """

        lut    = self.__selectedLut
        value  = lut.max() + 1
        name   = strings.labels['LutLabelDialog.newLabel']
        colour = fslcmaps.randomBrightColour()
        colour = [int(round(c * 255.0)) for c in colour]

        dlg = LutLabelDialog(self.GetTopLevelParent(), value, name, colour)
        if dlg.ShowModal() != wx.ID_OK:
            return

        lut    = self.__selectedLut
        value  = dlg.GetValue()
        name   = dlg.GetName()
        colour = dlg.GetColour()[:3]
        colour = [c / 255.0 for c in colour]

        if lut.get(value) is not None:
            wx.MessageBox(
                strings.messages[self, 'labelExists'].format(lut.name, value),
                strings.titles[  self, 'labelExists'], (wx.ICON_INFORMATION |
                                                        wx.OK))
            return

        log.debug('New lut label for {}: {}, {}, {}'.format(
            lut.name,
            value,
            name,
            colour))

        with lut.skip(self.name, 'added'):
            label  = lut.insert(value, name=name, colour=colour)
            widget = LabelWidget(self, lut, label)
            idx    = lut.index(label)
            self.__labelList.Insert(str(label.value),
                                    idx,
                                    clientData=label.value,
                                    extraWidget=widget)


    def __onLabelRemove(self, ev):
        """Called when the user pushes the *remove* button on the lookup
        table label list. Removes the selected label from the current
        :class:`.LookupTable`.
        """

        idx   = self.__labelList.GetSelection()
        lut   = self.__selectedLut
        value = lut[idx].value

        with lut.skip(self.name, 'removed'):
            lut.delete(value)
        self.__labelList.Delete(idx)


    def __selectedOverlayChanged(self, *a):
        """Called when the :class:`OverlayList` or
        :attr:`.DisplayContext.selectedOverlay` changes.  Refreshes the
        ``LookupTablePanel`` accordingly.
        """

        newOverlay = self.displayCtx.getSelectedOverlay()

        if self.__selectedOverlay is not None and \
           self.__selectedOverlay in self.overlayList:

            display = self.displayCtx.getDisplay(self.__selectedOverlay)
            display.removeListener('overlayType', self.name)

        self.__selectedOverlay = newOverlay

        if newOverlay is not None:

            display = self.displayCtx.getDisplay(newOverlay)
            display.addListener('overlayType',
                                self.name,
                                self.__overlayTypeChanged)

        self.__overlayTypeChanged()


    def __overlayTypeChanged(self, *a):
        """Called when the :attr:`.Display.overlayType` property of the
        currently selected overlay changes. If the :class:`.DisplayOpts`
        instance associated with the new overlay type is a
        :class:`.LabelOpts`, a listener is addd to its ``lut`` property.
        """

        if self.__selectedOpts is not None:
            self.__selectedOpts.removeListener('lut', self.name)
            self.__selectedOpts = None

        overlay = self.__selectedOverlay
        opts    = None

        if overlay is None:
            return

        opts = self.displayCtx.getOpts(overlay)

        if not isinstance(opts, (displayctx.LabelOpts, displayctx.MeshOpts)):

            # If the image is a Melodic image, show
            # the melodic classification lut
            if isinstance(overlay, fslmelimage.MelodicImage):
                self.__setLut(fslcmaps.getLookupTable('melodic-classes'))

            return

        opts.addListener('lut', self.name, self.__lutChanged)

        self.__selectedOpts = opts
        self.__lutChanged()
        self.Layout()


    def __lutChanged(self, *a):
        """Called when the :attr:`.LabelOpts.lut` property associated
        with the currently selected overlay changes. Changes the
        :class:`.LookupTable` displayed  on this ``LookupTablePanel`` (see
        the  :meth:`__setLut` method).
        """
        self.__setLut(self.__selectedOpts.lut)


def promptForLutName(initial=None):
    """Prompts the user to enter a name for a newly created
    :class:`.LookupTable`.
    """

    if initial is None:
        initial = strings.labels[LookupTablePanel, 'newLutDefault']

    nameMsg   = strings.messages[LookupTablePanel, 'newlut']
    nameTitle = strings.titles[  LookupTablePanel, 'newlut']

    while True:
        dlg = wx.TextEntryDialog(
            wx.GetApp().GetTopWindow(),
            nameMsg,
            nameTitle,
            initial)

        if dlg.ShowModal() != wx.ID_OK:
            return None, None

        lutName = dlg.GetValue()
        lutKey  = fslcmaps.makeValidMapKey(lutName)

        # a lut with the specified name already exists
        if fslcmaps.isLookupTableRegistered(lutKey):
            nameMsg = strings.messages[LookupTablePanel, 'alreadyinstalled']
            continue

        break

    return lutKey, lutName


class LabelWidget(wx.Panel):
    """A ``LabelWidget`` is shown for each label of the :class:`.LookupTable`
    which is currently displayed on a :class:`LookupTablePanel`. A
    ``LabelWidget`` allows the user to change the colour and visibility of the
    label value.
    """


    def __init__(self, lutPanel, lut, label):
        """Create a ``LabelWidget``.

        :arg lutPanel: The :class:`LookupTablePanel` that is displaying this
                       ``LabelWidget``.

        :arg lut:      The :class:`.LookupTable` currently being displayed.

        :arg label:    The :class:`.LutLabel` that this ``LabelWidget`` is
                       associated with.
        """
        wx.Panel.__init__(self, lutPanel)

        self.__lutPanel = lutPanel
        self.__lut      = lut
        self.__label    = label

        self.__name   = props.makeWidget(self, label, 'name')
        self.__enable = props.makeWidget(self, label, 'enabled')
        self.__colour = props.makeWidget(self, label, 'colour')

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(self.__sizer)
        self.__sizer.Add(self.__enable, flag=wx.ALIGN_CENTRE)
        self.__sizer.Add(self.__colour, flag=wx.ALIGN_CENTRE)
        self.__sizer.Add(self.__name, flag=wx.EXPAND)


class LutLabelDialog(wx.Dialog):
    """A dialog which is displayed when the user adds a new label to the
    :class:`.LookupTable` currently displayed in the
    :class:`LookupTablePanel`.

    Prompts the user to enter a label value, name, and colour. After the
    dialog is dismissed, the entered information is available via the
    following methods:

    .. autosummary::
       :nosignatures:

       GetValue
       GetName
       GetColour
    """


    def __init__(self, parent, value, name, colour):
        """Create a ``LutLabelDialog``.

        :arg parent: The :mod:`wx` parent object.
        """

        wx.Dialog.__init__(self, parent, title=strings.titles[self])

        self.__value  = wx.SpinCtrl(        self, min=0, max=65535)
        self.__name   = wx.TextCtrl(        self)
        self.__colour = wx.ColourPickerCtrl(self)

        self.__valueLabel  = wx.StaticText(self)
        self.__nameLabel   = wx.StaticText(self)
        self.__colourLabel = wx.StaticText(self)

        self.__ok     = wx.Button(self, id=wx.ID_OK)
        self.__cancel = wx.Button(self, id=wx.ID_CANCEL)

        self.__valueLabel .SetLabel(strings.labels[self, 'value'])
        self.__nameLabel  .SetLabel(strings.labels[self, 'name'])
        self.__colourLabel.SetLabel(strings.labels[self, 'colour'])
        self.__ok         .SetLabel(strings.labels[self, 'ok'])
        self.__cancel     .SetLabel(strings.labels[self, 'cancel'])

        self.__value      .SetValue( value)
        self.__name       .SetValue( name)
        self.__colour     .SetColour(colour)

        self.__sizer = wx.GridSizer(4, 2, 0, 0)
        self.SetSizer(self.__sizer)

        self.__sizer.Add(self.__valueLabel,
                         flag=wx.EXPAND | wx.TOP | wx.LEFT,
                         border=15)
        self.__sizer.Add(self.__value,
                         flag=wx.EXPAND | wx.TOP | wx.RIGHT,
                         border=15)
        self.__sizer.Add(self.__nameLabel,
                         flag=wx.EXPAND | wx.LEFT,
                         border=15)
        self.__sizer.Add(self.__name,
                         flag=wx.EXPAND | wx.RIGHT,
                         border=15)
        self.__sizer.Add(self.__colourLabel,
                         flag=wx.EXPAND | wx.LEFT,
                         border=15)
        self.__sizer.Add(self.__colour,
                         flag=wx.EXPAND | wx.RIGHT,
                         border=15)
        self.__sizer.Add(self.__ok,
                         flag=wx.EXPAND | wx.BOTTOM | wx.LEFT,
                         border=15)
        self.__sizer.Add(self.__cancel,
                         flag=wx.EXPAND | wx.BOTTOM | wx.RIGHT,
                         border=15)

        self.__ok    .Bind(wx.EVT_BUTTON, self.__onOk)
        self.__cancel.Bind(wx.EVT_BUTTON, self.__onCancel)

        self.__ok.SetDefault()

        self.Layout()
        self.Fit()

        self.CentreOnParent()

        self.__enteredValue  = None
        self.__enteredName   = None
        self.__enteredColour = None


    def GetValue(self):
        """Returns the value that was entered by the user. Or, returns
        ``None`` if  the user cancelled the dialog, or the dialog has not
        yet been closed.
        """
        return self.__enteredValue


    def GetName(self):
        """Returns the name that was entered by the user. Or, returns
        ``None`` if  the user cancelled the dialog, or the dialog has not
        yet been closed.
        """
        return self.__enteredName


    def GetColour(self):
        """Returns the colour that was entered by the user. Or, returns
        ``None`` if  the user cancelled the dialog, or the dialog has not
        yet been closed.
        """
        return self.__enteredColour


    def __onOk(self, ev):
        """Called when the user confirms the dialog. Saves the name, colour,
        and value that were entered, and closes the dialog.
        """
        self.__enteredValue  = self.__value .GetValue()
        self.__enteredName   = self.__name  .GetValue()
        self.__enteredColour = self.__colour.GetColour()

        self.EndModal(wx.ID_OK)


    def __onCancel(self, ev):
        """Called when the user cancells the dialog. Closes the dialog."""

        self.EndModal(wx.ID_CANCEL)



#!/usr/bin/env python
#
# scene3dviewprofile.py - The Scene3DViewProfile class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Scene3DViewProfile` class, an interaction
:class:`.Profile` for :class:`.Scene3DPanel` views.
"""


import logging

import          wx
import numpy as np

import fsleyes_props        as props
import fsl.transform.affine as affine
import fsl.utils.idle       as idle
import fsleyes.profiles     as profiles
import fsleyes.actions      as actions


log = logging.getLogger(__name__)


class Scene3DViewProfile(profiles.Profile):
    """The :class:`Scene3DViewProfile` class is a :class:`.Profile` for
    :class:`.Scene3DPanel` views. It defines mouse / keyboard handlers for
    interacting with the :class:`.Scene3DCanvas` contained in the panel.

    The following *modes* are defined (see the :class:`.Profile`
    documentation):

    ========== ========================================================
    ``rotate`` Clicking and dragging the mouse rotates the scene
    ``zoom``   Moving the mouse wheel zooms in and out.
    ``pan``    Clicking and dragging the mouse pans the scene.
    ``pick``   Clicking changes the :attr:`.DisplayContext.location`
    ========== ========================================================
    """


    @staticmethod
    def supportedView():
        """Specifies that this profile can only work with the
        :class:`.Scene3DPanel` view.
        """
        from fsleyes.views.scene3dpanel import Scene3DPanel
        return Scene3DPanel


    @staticmethod
    def tempModes():
        """Returns the temporary mode map for the ``Scene3DViewProfile``,
        which controls the use of modifier keys to temporarily enter other
        interaction modes.
        """
        return {
            ('rotate', wx.WXK_CONTROL) : 'zoom',
            ('rotate', wx.WXK_ALT)     : 'pan',
            ('rotate', wx.WXK_SHIFT)   : 'pick'}


    @staticmethod
    def altHandlers():
        """Returns the alternate handlers map, which allows event handlers
        defined in one mode to be re-used whilst in another mode.
        """
        return {
            ('rotate', 'MiddleMouseDown') : ('pan', 'LeftMouseDown'),
            ('rotate', 'MiddleMouseDrag') : ('pan', 'LeftMouseDrag'),
            ('rotate', 'MiddleMouseUp')   : ('pan', 'LeftMouseUp')}


    def __init__(self,
                 viewPanel,
                 overlayList,
                 displayCtx):

        modes = ['rotate', 'zoom', 'pan', 'pick']

        profiles.Profile.__init__(self,
                                  viewPanel,
                                  overlayList,
                                  displayCtx,
                                  modes)

        self.__canvas         = viewPanel.getGLCanvases()[0]
        self.__rotateMousePos = None
        self.__panMousePos    = None
        self.__panCanvasPos   = None


    def getEventTargets(self):
        """Returns a list containing the :class:`.Scene3DCanvas`.
        """
        return [self.__canvas]


    @actions.action
    def resetDisplay(self):
        """Resets the :class:`.Scene3DCanvas` camera settings to their
        defaults.
        """

        opts = self.__canvas.opts

        with props.suppressAll(opts):
            opts.zoom     = 75
            opts.offset   = [0, 0]
            opts.rotation = np.eye(3)
        self.__canvas.Refresh()


    def _rotateModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Called on left mouse down events in ``rotate`` mode.
        Saves the mouse position and current rotation matrix (the
        :attr:`.Scene3DCanvas.rotation` property).
        """
        self.__rotateMousePos = mousePos
        self.__baseXform      = canvas.opts.rotation
        self.__lastRot        = np.eye(3)


    def _rotateModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Called on left mouse drag events in ``rotate`` mode.
        Modifies the canvas rotation matrix according to the X and Y
        mouse position (relative to the mouse down location).
        """

        if self.__rotateMousePos is None:
            return

        w, h   = canvas.GetSize()
        x0, y0 = self.__rotateMousePos
        x1, y1 = mousePos


        # Normalise x/y mouse pos to [-fac*pi, +fac*pi]
        fac = 1
        x0  = -1 + 2 * (x0 / float(w)) * fac * np.pi
        y0  = -1 + 2 * (y0 / float(h)) * fac * np.pi
        x1  = -1 + 2 * (x1 / float(w)) * fac * np.pi
        y1  = -1 + 2 * (y1 / float(h)) * fac * np.pi

        xrot = x1 - x0
        yrot = y1 - y0

        rot = affine.axisAnglesToRotMat(yrot, 0, xrot)

        self.__lastRot        = rot
        self.__rotateMousePos = mousePos

        canvas.opts.rotation = affine.concat(rot,
                                                self.__lastRot,
                                                self.__baseXform)


    def _rotateModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Called on left mouse up events in ``rotate`` mode.
        Clears the internal state used by the mouse down and drag
        handlers.
        """
        self.__rotateMousePos = None


    def _zoomModeMouseWheel(self, ev, canvas, wheel, mousePos, canvasPos):
        """Called on mouse wheel events in ``zoom`` mode. Adjusts the
        :attr:`.Scene3DCanvas.zoom` property.
        """

        if wheel == 0:
            return

        opts = canvas.opts

        def update():
            if   wheel > 0: opts.zoom += 0.1 * opts.zoom
            elif wheel < 0: opts.zoom -= 0.1 * opts.zoom

        # See comment in OrthoViewProfile._zoomModeMouseWheel
        # for the reason why we do this asynchronously.
        idle.idle(update, timeout=0.1)


    def _panModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Called on mouse down events in ``pan`` mode. Saves the
        mouse position and current :attr:`.Scene3DCanvas.offset`
        value.
        """
        x, y = mousePos
        w, h = canvas.GetSize()
        x    = -1 + 2 * x / float(w)
        y    = -1 + 2 * y / float(h)

        self.__panMousePos    = (x, y)
        self.__panStartOffset = canvas.opts.offset[:]


    def _panModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Called on mouse drag events in ``pan`` mode. Adjusts the
        :attr:`.Scene3DCanvas.offset` property.
        """
        if self.__panMousePos is None:
            return

        w,  h  = canvas.GetSize()
        sx, sy = self.__panMousePos
        ox, oy = self.__panStartOffset
        ex, ey = mousePos
        ex     = -1 + 2 * ex / float(w)
        ey     = -1 + 2 * ey / float(h)

        canvas.opts.offset = [ox + ex - sx, oy + ey - sy]


    def _panModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Called on mouse up events in ``pan`` mode. Clears the
        internal state used by the down and drag handlers.
        """
        self.__panMousePos  = None


    def _pickModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Called on mouse down events in ``pick`` mode.

        Updates the :attr:`DisplayContext.location` property.
        """

        from fsl.data.mesh import Mesh

        displayCtx = self.displayCtx
        ovl        = displayCtx.getSelectedOverlay()

        if ovl is None:
            return

        # The canvasPos is located on the near clipping
        # plane (see Scene3DCanvas.canvasToWorld).
        # We also need the corresponding point on the
        # far clipping plane.
        farPos = canvas.canvasToWorld(mousePos[0], mousePos[1], near=False)

        # For non-mesh overlays, we select a point which
        # is in between the near/far clipping planes.
        if not isinstance(ovl, Mesh):

            posDir = farPos - canvasPos
            dist   = affine.veclength(posDir)
            posDir = affine.normalise(posDir)
            midPos = canvasPos + 0.5 * dist * posDir

            self.displayCtx.location.xyz = midPos

        else:
            opts      = self.displayCtx.getOpts(ovl)
            rayOrigin = canvasPos
            rayDir    = affine.normalise(farPos - canvasPos)

            # transform location from display into model space
            rayOrigin = opts.transformCoords(rayOrigin, 'display', 'mesh')
            rayDir    = opts.transformCoords(rayDir,    'display', 'mesh',
                                             vector=True)
            loc, tri  = ovl.rayIntersection([rayOrigin],
                                            [rayDir],
                                            vertices=True)

            if len(loc) == 0:
                return

            loc = loc[0]
            tri = ovl.indices[int(tri[0]), :]

            # The rayIntersection method gives us a
            # point on one of the mesh triangles -
            # we want the vertex on that triangle
            # which is nearest to the intersection.
            triVerts = ovl.vertices[tri, :]
            triDists = affine.veclength(loc - triVerts)
            vertIdx  = np.argsort(triDists)[0]

            loc      = ovl.vertices[tri[vertIdx], :]
            loc      = opts.transformCoords(loc, 'mesh', 'display')

            self.displayCtx.location.xyz = loc


    def _pickModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Called on mouse drag events in ``pick`` mode. Forwards the
        event to the :meth:`_pickModeLeftMouseDown` method.
        """
        self._pickModeLeftMouseDown(ev, canvas, mousePos, canvasPos)



#!/usr/bin/env python
#
# applycommandline.py - The ApplyCommandLineAction
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ApplyCommandLineAction` class, an
:class:`.Action` which allows the user to apply FSLeyes command line
arguments to a :class:`.CanvasPanel`. The stand-alone
:func:`applyCommandLineArgs` function is where the work is actually
implemented.
"""


import os.path as op
import            os
import            sys
import            argparse
import            contextlib
import            six

import wx

import fsleyes_widgets.dialog       as fsldlg
import fsleyes_widgets.utils.status as status

import fsleyes.strings              as strings
import fsleyes.parseargs            as parseargs
from . import                          base


class ApplyCommandLineAction(base.Action):
    """The :class:`ApplyCommandLineAction` class is an :class:`.Action` which
    allows the user to apply FSLeyes command line arguments to a
    :class:`.CanvasPanel`.
    """


    def __init__(self, overlayList, displayCtx, panel):
        """Create an ``ApplyCommandLineAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg panel:       The :class:`.CanvasPanel`.
        """
        base.Action.__init__(
            self, overlayList, displayCtx, self.__applyCommandLineArgs)
        self.__panel = panel


    def __applyCommandLineArgs(self):
        """Called when this action is executed. Prompts the user to enter
        some command line arguments, and then passes them to the
        :func:`applyCommandLineArgs` function.
        """

        # prompt user for some arguments
        dlg = fsldlg.TextEditDialog(
            self.__panel,
            strings.titles[  self, 'title'],
            strings.messages[self, 'apply'],
            style=fsldlg.TED_MULTILINE | fsldlg.TED_OK_CANCEL)

        dlg.CentreOnParent()

        if dlg.ShowModal() != wx.ID_OK:
            return

        # Pass GetText through str, because it
        # returns a unicode string, and python's
        # argparse does not seem to like unicode.
        argv     = str(dlg.GetText()).split()
        errTitle = strings.titles[  self, 'error']
        errMsg   = strings.messages[self, 'error']

        # apply said arguments
        with status.reportIfError(errTitle, errMsg):
            applyCommandLineArgs(self.overlayList,
                                 self.displayCtx,
                                 argv,
                                 self.__panel)


class ApplyCLIExit(Exception):
    """``Exception`` class raised by the :func:`applyCommandLineArgs`
    function.
    """

    def __init__(self, code, stdout, stderr):
        self.code   = code
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        return '\n'.join((self.stderr, self.stdout))


@contextlib.contextmanager
def chdir(dest=None):
    """Context manager which temporarily changes to the specified ``dest``
    directory, then restores the current directory afterwards.

    If ``dest is None``, the function yields without doing anything.
    """

    if dest is None:
        yield
        return

    src = os.getcwd()
    os.chdir(dest)

    try:
        yield
    finally:
        os.chdir(src)


def applyCommandLineArgs(overlayList,
                         displayCtx,
                         argv,
                         panel=None,
                         applyOverlayArgs=True,
                         baseDir=None,
                         **kwargs):
    """Applies the command line arguments stored in ``argv`` to the
    :class:`.CanvasPanel` ``panel``. If ``panel is None``, it is assumed
    that ``argv`` only contains overlay arguments.

    :arg overlayList:      The :class:`.OverlayList`.

    :arg displayCtx:       The :class:`.DisplayContext`. If a ``panel`` is
                           provided, this should be the ``DisplayContext``
                           associated with that panel.

    :arg argv:             List of command line arguments to apply.

    :arg panel:            Optional :class:`.CanvasPanel` to apply the
                           arguments to.

    :arg applyOverlayArgs: If ``False``, overlay arguments are not applied.

    :arg baseDir:          Directory from which to interpret the arguments,
                           in case this is different from the current working
                           directory, and overlays have been specified with
                           relative paths.

    All other keyword arguments are passed to the
    :func:`.parseargs.applyOverlayArgs`  function.
    """

    # We patch sys.stdout/stderr
    # while parseargs.parseArgs is
    # called so we can capture its
    # output.
    stdout = six.StringIO()
    stderr = six.StringIO()

    if argv[0] == 'fsleyes':
        argv = argv[1:]

    parser = argparse.ArgumentParser(add_help=False)

    try:
        real_stdout = sys.stdout
        real_stderr = sys.stderr
        sys.stdout  = stdout
        sys.stderr  = stderr

        with chdir(baseDir):
            namespace = parseargs.parseArgs(parser, argv, 'fsleyes')

    except SystemExit as e:
        raise ApplyCLIExit(e.code, stdout.getvalue(), stderr.getvalue())

    finally:
        sys.stdout = real_stdout
        sys.stderr = real_stderr

    if baseDir is not None:
        for o in namespace.overlays:
            if not op.isabs(o.overlay):
                o.overlay = op.join(baseDir, o.overlay)

    if applyOverlayArgs:
        parseargs.applyOverlayArgs(
            namespace, overlayList, displayCtx, **kwargs)

    if panel is not None:
        sceneOpts = panel.sceneOpts
        parseargs.applySceneArgs(namespace, overlayList, displayCtx, sceneOpts)



#!/usr/bin/env python
#
# labelopts.py - The LabelOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LabelOpts` class, which defines settings
for displaying :class:`.Image` overlays as label images., such as anatomical
atlas images, tissue segmentation images, and so on.
"""


import fsleyes_props      as props
import fsleyes.colourmaps as colourmaps
from . import                niftiopts


class LabelOpts(niftiopts.NiftiOpts):
    """The ``LabelOpts`` class defines settings for displaying
    :class:`.Image` overlays as label images., such as anatomical atlas
    images, tissue segmentation images, and so on.
    """

    lut = props.Choice()
    """The :class:`.LookupTable` used to colour each label.
    """

    outline = props.Boolean(default=False)
    """If ``True`` only the outline of contiguous regions with the same label
    value will be shown. If ``False``, contiguous regions will be filled.
    """


    outlineWidth = props.Int(minval=0, maxval=10, default=1, clamped=True)
    """Width of labelled region outlines, if :attr:``outline` is ``True``.
    This value is in terms of pixels.
    """


    showNames = props.Boolean(default=False)
    """If ``True``, region names (as defined by the current
    :class:`.LookupTable`) will be shown alongside each labelled region.

    .. note:: Not implemented yet.
    """


    def __init__(self, overlay, *args, **kwargs):
        """Create a ``LabelOpts`` instance for the specified ``overlay``.
        All arguments are passed through to the :class:`.NiftiOpts`
        constructor.
        """

        # Some FSL tools will set the nifti aux_file
        # field to the name of a colour map - Check
        # to see if this is the case (again, before
        # calling __init__, so we don't clobber any
        # existing values).
        aux_file = overlay.strval('aux_file').lower()

        if aux_file.startswith('mgh'):
            aux_file = 'freesurfercolorlut'

        # Check to see if any registered lookup table
        # has an ID that starts with the aux_file value.
        # Default to random lut if aux_file is empty,
        # or does not correspond to a registered lut.
        lut = 'random'

        if aux_file != '':
            luts = colourmaps.getLookupTables()
            luts = [l.key for l in luts if l.key.startswith(aux_file)]

            if len(luts) == 1:
                lut = luts[0]

        self.lut = lut

        niftiopts.NiftiOpts.__init__(self, overlay, *args, **kwargs)



#!/usr/bin/env python
#
# glmip_funcs.py - Functions used by GLMIP for rendering in an OpenGL 2.1
# environment.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions used by the :class:`.GLMIP` class for
rendering in an OpenGL 2.1 environment.
"""

import numpy as np

import fsl.transform.affine as affine
import fsleyes.gl.shaders   as shaders
from . import                  glvolume_funcs


def init(self):
    """Initialise the shader programs. """
    self.shader = None
    compileShaders(   self)
    updateShaderState(self)


def destroy(self):
    """Destroy the shader programs. """
    self.shader.destroy()
    self.shader = None


def compileShaders(self):
    """Compiles vertex and fragment shaders. """
    if self.shader is not None:
        self.shader.destroy()

    vertSrc = shaders.getVertexShader(  'glvolume')
    fragSrc = shaders.getFragmentShader('glmip')

    self.shader = shaders.GLSLShader(vertSrc, fragSrc)


def updateShaderState(self):
    """Updates the vertex/fragment shader state based on the current
    state of the :class:`.MIPOpts` instance.
    """

    if not self.ready():
        return

    opts   = self.opts
    shader = self.shader

    vmin, vmax = self.overlay.dataRange

    # Convert clipping values from voxel value
    # range totexture value range (0.0 - 1.0).
    imgXform   = self.imageTexture.invVoxValXform
    clipLow    = opts.clippingRange[0] * imgXform[0, 0] + imgXform[0, 3]
    clipHigh   = opts.clippingRange[1] * imgXform[0, 0] + imgXform[0, 3]
    textureMin = vmin                  * imgXform[0, 0] + imgXform[0, 3]
    textureMax = vmax                  * imgXform[0, 0] + imgXform[0, 3]
    imageShape = self.image.shape[:3]

    # Create a single transformation matrix
    # which transforms from image texture values
    # to voxel values, and scales said voxel
    # values to colour map texture coordinates.
    img2CmapXform = affine.concat(
        self.cmapTexture.getCoordinateTransform(),
        self.imageTexture.voxValXform)

    # sqrt(3) so the window is 100%
    # along the diagonal of a cube
    window = np.sqrt(3) * opts.window / 100.0

    shader.load()

    changed = False

    changed |= shader.set('imageTexture',     0)
    changed |= shader.set('cmapTexture',      1)
    changed |= shader.set('textureMin',       textureMin)
    changed |= shader.set('textureMax',       textureMax)
    changed |= shader.set('img2CmapXform',    img2CmapXform)
    changed |= shader.set('imageShape',       imageShape)
    changed |= shader.set('useSpline',        opts.interpolation == 'spline')
    changed |= shader.set('clipLow',          clipLow)
    changed |= shader.set('clipHigh',         clipHigh)
    changed |= shader.set('invertClip',       opts.invertClipping)
    changed |= shader.set('window',           window)
    changed |= shader.set('useMinimum',       opts.minimum)
    changed |= shader.set('useAbsolute',      opts.absolute)

    shader.unload()

    return changed


def draw2D(self, zpos, axes, xform=None, bbox=None):
    """Draws a 2D slice at the given ``zpos``. Uses the
    :func:`.gl21.glvolume_funcs.draw2D` function.
    """
    self.shader.load()

    viewmat       = self.canvas.viewMatrix
    cdir, rayStep = self.opts.calculateRayCastSettings(viewmat)

    self.shader.set('cameraDir', cdir)
    self.shader.set('rayStep',   rayStep)

    glvolume_funcs.draw2D(self, zpos, axes, xform, bbox)
    self.shader.unloadAtts()
    self.shader.unload()



#!/usr/bin/env python
#
# __init__.py - GLSL shader program management.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``glsl`` package is the home of the :class:`.GLSLShader` class, a class
which may be used to manage GLSL shader programs. The :class:`.GLSLShader`
class is defined in the :mod:`.program` module, and uses some functions
defined in the :mod:`.parse` module.

See also the :mod:`.arbp` package.
"""



#!/usr/bin/env python
#
# glrgbvector_funcs.py - OpenGL 1.4 functions used by the GLRGBVector class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions which are used by the :class:`.GLRGBVector`
class to render :class:`.Image` overlays as RGB vector images in an OpenGL 1.4
compatible manner.


This module uses functions in the :mod:`.gl14.glvector_funcs` module, which
contains logic used for rendering both ``GLRGBVector`` and ``GLLineVector``
instances.


Rendering of a ``GLRGBVector`` is very similar to that of a
:class:`.GLVolume`, with the exception that a different fragment shader
(``glvector``) may be used. Therefore, the ``preDraw``, ``draw2D``,
``draw3D``, ``drawAll`` and ``postDraw`` functions defined in the
:mod:`.gl14.glvolume_funcs` are re-used by this module.
"""


from . import glvolume_funcs
from . import glvector_funcs


def init(self):
    """Calls the :func:`compileShaders` and :func:`updateShaderState`
    functions.
    """

    self.shader = None

    compileShaders(   self)
    updateShaderState(self)


def destroy(self):
    """Destroys the vertex/fragment shader programs created in :func:`init`.
    """
    glvector_funcs.destroy(self)


def compileShaders(self):
    """Calls the :mod:`.gl14.glvector_funcs.compileShaders` function.
    """
    glvector_funcs.compileShaders(self, 'glrgbvector')


def updateShaderState(self):
    """Updates all variables used by the vertex/fragment shader programs.  The
    fragment shader is configured by the
    :func:.gl21.glvector_funcs.updateFragmentShaderState` function.
    """

    if not self.ready():
        return

    self.shader.load()
    glvector_funcs.updateShaderState(self)
    self.shader.unload()

    return True


preDraw  = glvolume_funcs.preDraw
draw2D   = glvolume_funcs.draw2D
draw3D   = glvolume_funcs.draw3D
drawAll  = glvolume_funcs.drawAll
postDraw = glvolume_funcs.postDraw



#!/usr/bin/env python
#
# version.py - FSLeyes version information.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The sole purpose of this module is as a container for the *FSLeyes*
version number. See also the :mod:`fsl.version` module.

.. autosummary::
   :nosignatures:

   __version__
"""


__version__ = '1.0.13'
"""Current version number, as a string. The FSLeyes version number consists
of three numbers, separated by a period, which roughly obeys the Semantic
Versioning conventions (http://semver.org/).
"""



#!/usr/bin/env python
#
# colourbarcanvas.py - Render a colour bar using OpenGL and matplotlib.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ColourBarCanvas`.

The :class:`ColourBarCanvas` uses a :class:`.ColourBar` draw a colour bar
(with labels), and then renders said colour bar as a texture using OpenGL.

See the :mod:`~fsleyes.controls.colourbar` and
:mod:`fsleyes_widgets.utils.colourbarbitmap` modules for details on how the
colour bar is created.
"""


import logging

import numpy     as np
import OpenGL.GL as gl

import fsleyes_props                         as props
import fsl.utils.idle                        as idle
import fsleyes.controls.colourbar            as cbar
import fsleyes.gl.textures                   as textures


log = logging.getLogger(__name__)


class ColourBarCanvas(props.HasProperties):
    """Contains logic to render a colour bar as an OpenGL texture. """


    highDpi = props.Boolean(default=False)
    """Scale colour bar canvas for high-resolution screens. """


    barSize = props.Percentage(default=100)
    """Size of the colour bar along its major axis, as a proportion of
    the available space.
    """


    def __init__(self, overlayList, displayCtx):
        """Adds a few listeners to the properties of this object, to update
        the colour bar when they change.
        """

        self.__tex  = None
        self.__name = '{}_{}'.format(self.__class__.__name__, id(self))
        self.__cbar = cbar.ColourBar(overlayList, displayCtx)
        self.__cbar.register(self.__name, self.updateColourBarTexture)

        self.addListener('barSize', self.__name, self.updateColourBarTexture)
        self.addListener('highDpi', self.__name, self.__highDpiChanged)


    @property
    def colourBar(self):
        """Returns a reference to the :class:`.ColourBar` object that actually
        generates the colour bar bitmap.
        """
        return self.__cbar


    def updateColourBarTexture(self, *a):
        """Called whenever the colour bar texture needs to be updated. """

        def update():
            self.__genColourBarTexture()
            self.Refresh()

        name = '{}_updateColourBarTexture'.format(id(self))

        idle.idle(update, name=name, skipIfQueued=True)


    def _initGL(self):
        """Called automatically by the OpenGL canvas target superclass (see the
        :class:`.WXGLCanvasTarget` and :class:`.OSMesaCanvasTarget` for
        details).

        Generates the colour bar texture.
        """
        self.__genColourBarTexture()


    def __highDpiChanged(self, *a):
        """Called when the :attr:`highDpi` property changes. Calls the
        :meth:`.GLCanvasTarget.EnableHighDPI` method.
        """
        self.EnableHighDPI(self.highDpi)
        self.updateColourBarTexture()


    def destroy(self):
        """Should be called when this ``ColourBarCanvas`` is no longer needed.
        Destroys the :class:`.Texture2D` and :class:`.ColourBar` instances
        used to render the colour bar.
        """
        self.__cbar.deregister(self.__name)
        self.__cbar.destroy()

        if self.__tex is not None:
            self.__tex.destroy()

        self.removeListener('barSize', self.__name)
        self.removeListener('highDpi', self.__name)

        self.__tex  = None
        self.__cbar = None


    def __genColourBarTexture(self):
        """Retrieves a colour bar bitmap from the :class:`.ColourBar`, and
        copies it to a :class:`.Texture2D`.
        """

        if not self._setGLContext():
            return

        # we may have already
        # been destroyed
        if self.__cbar is None:
            return

        w, h = self.GetSize()

        if w < 50 or h < 50:
            return

        if self.__cbar.orientation == 'vertical': h = h * self.barSize / 100.0
        else:                                     w = w * self.barSize / 100.0

        scale  = self.GetScale()
        bitmap = self.__cbar.colourBar(w, h, scale)

        # The bitmap has shape W*H*4, but
        # Texture2D instances need it in
        # shape 4*W*H
        if bitmap is None:
            return

        bitmap = np.fliplr(bitmap).transpose([2, 0, 1])

        if self.__tex is None:
            self.__tex = textures.Texture2D('{}_{}'.format(
                type(self).__name__, id(self)), interp=gl.GL_LINEAR)

        self.__tex.set(data=bitmap)


    def _draw(self):
        """Renders the colour bar texture using all available canvas space."""

        if self.__tex is None or not self._setGLContext():
            return

        width, height = self.GetScaledSize()

        # viewport
        gl.glViewport(0, 0, width, height)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        gl.glOrtho(0, 1, 0, 1, -1, 1)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()

        gl.glClearColor(*self.__cbar.bgColour)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
        gl.glShadeModel(gl.GL_FLAT)

        xmin, xmax = 0, 1
        ymin, ymax = 0, 1
        off        = (100 - self.barSize) / 100.0

        if self.colourBar.orientation == 'vertical':
            ymin += off / 2.0
            ymax -= off / 2.0
        else:
            xmin += off / 2.0
            xmax -= off / 2.0

        self.__tex.drawOnBounds(0, xmin, xmax, ymin, ymax, 0, 1)



#!/usr/bin/env python
#
# dataseries.py - The DataSeries class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`DataSeries` class, the base class for
classes used by the :class:`.PlotCanvas` for plotting data.
"""


import logging

import numpy as np

import fsl.utils.idle  as idle
import fsl.utils.cache as cache
import fsleyes_props   as props


log = logging.getLogger(__name__)


class DataSeries(props.HasProperties):
    """A ``DataSeries`` instance encapsulates some data to be plotted by a
    :class:`PlotCanvas`, possibly with the data extracted from an overlay in
    the :class:`.OverlayList`.

    Sub-class implementations must:

      - Accept an overlay object, :class:`.OverlayList`,
        :class:`.DisplayContext`, and :class:`.PlotCanvas` in their
        ``__init__`` method, and pass these through to
        :meth:`.DataSeries.__init__`.
      - Override the :meth:`getData` method
      - Override the :meth:`redrawProperties` method if necessary


    The overlay is accessible as an instance attribute called ``overlay``.


    .. note:: Some ``DataSeries`` instances may not be associated with
              an overlay (e.g. series imported loaded a text file). In
              this case, the ``overlay`` attribute will be ``None``.


    Each``DataSeries`` instance is plotted as a line, with the line
    style defined by properties on the ``DataSeries`` instance,
    such as :attr:`colour`, :attr:`lineWidth` etc.
    """


    colour = props.Colour()
    """Line colour. """


    enabled = props.Boolean(default=True)
    """Draw or not draw?"""


    alpha = props.Real(minval=0.0, maxval=1.0, default=1.0, clamped=True)
    """Line transparency."""


    label = props.String()
    """Line label (used in the plot legend)."""


    lineWidth = props.Choice((0.5, 1, 2, 3, 4, 5))
    """Line width. """


    lineStyle = props.Choice(('-',
                              '--',
                              '-.',
                              ':',
                              (0, (5, 7)),
                              (0, (1, 7)),
                              (0, (4, 10, 1, 10)),
                              (0, (4, 1, 1, 1, 1, 1)),
                              (0, (4, 1, 4, 1, 1, 1))))
    """Line style. See
    https://matplotlib.org/gallery/lines_bars_and_markers/linestyles.html
    """


    def __init__(self, overlay, overlayList, displayCtx, plotCanvas):
        """Create a ``DataSeries``.

        :arg overlay:     The overlay from which the data to be plotted is
                          retrieved.  May be ``None``.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg plotCanvas:  The :class:`.PlotCanvas` that owns this
                          ``DataSeries``.
        """

        self.__name        = '{}_{}'.format(type(self).__name__, id(self))
        self.__overlay     = overlay
        self.__overlayList = overlayList
        self.__displayCtx  = displayCtx
        self.__plotCanvas  = plotCanvas
        self.setData([], [])

        log.debug('{}.init ({})'.format(type(self).__name__, id(self)))


    def __del__(self):
        """Prints a log message. """
        if log:
            log.debug('{}.del ({})'.format(type(self).__name__, id(self)))


    def __hash__(self):
        """Returns a hash for this ``DataSeries`` instance."""
        return hash(id(self))


    @property
    def name(self):
        """Returns a unique name for this ``DataSeries`` instance. """
        return self.__name


    @property
    def overlay(self):
        """Returns the overlay associated with this ``DataSeries`` instance.
        """
        return self.__overlay


    @property
    def overlayList(self):
        """Returns the :class:`.OverlayList`.
        """
        return self.__overlayList


    @property
    def displayCtx(self):
        """Returns the :class:`.DisplayContext`.
        """
        return self.__displayCtx


    @property
    def plotCanvas(self):
        """Returns the :class:`.PlotCanvas` that owns this ``DataSeries``
        instance.
        """
        return self.__plotCanvas


    def destroy(self):
        """This method must be called when this ``DataSeries`` instance is no
        longer needed. This implementation may be overridden by sub-classes
        which need to perform any clean-up operations. Sub-class
        implementations should call this implementation.
        """
        self.__overlay     = None
        self.__overlayList = None
        self.__displayCtx  = None
        self.__plotCanvas  = None


    def redrawProperties(self):
        """Returns a list of all properties which, when their values change,
        should result in this ``DataSeries`` being re-plotted. This method
        may be overridden by sub-classes.
        """

        return self.getAllProperties()[0]


    def extraSeries(self):
        """Some ``DataSeries`` types have additional ``DataSeries`` associated
        with them (see e.g. the :class:`.FEATTimeSeries` class). This method
        can be overridden to return a list of these extra ``DataSeries``
        instances. The default implementation returns an empty list.
        """
        return []


    def setData(self, xdata, ydata):
        """Set the data to be plotted. This method is irrelevant if a
        ``DataSeries`` sub-class has overridden :meth:`getData`.
        """
        self.__xdata = xdata
        self.__ydata = ydata


    def getData(self):
        """This method should be overridden by sub-classes. It must return
        the data to be plotted, as a tuple of the form:

            ``(xdata, ydata)``

        where ``xdata`` and ``ydata`` are sequences containing the x/y data
        to be plotted.

        The default implementation returns the data that has been set via the
        :meth:`setData` method.
        """
        return self.__xdata, self.__ydata


class VoxelDataSeries(DataSeries):
    """The ``VoxelDataSeries`` class is a :class:`DataSeries` class which
    provides some functionality useful to data series that represent data
    from a voxel in an :class:`.Image` overlay.

    It contains a built-in cache which is used to prevent repeated access
    to data from the same voxel.

    Sub-classes may need to override:

      - the :meth:`currentVoxelLocation` method, which
        generates the location index for the current location, and which
        is used as the unique cache key when the corresponding data is cached

      - The :meth:`currentVoxelData` method, which retrieves and/or calculates
        the data at the current location. This is the data which is cached,
        and which is returned by the :meth:`dataAtCurrentVoxel` method.
    """


    def __init__(self, *args, **kwargs):
        """Create a ``VoxelDataSeries``. All arguments are passed through to
        the :class:`DataSeries` constructor.
        """

        DataSeries.__init__(self, *args, **kwargs)

        # We use a cache to store data for the
        # most recently accessed voxels. This is
        # done to improve performance on big
        # images (which may be compressed and
        # on disk).
        #
        # TODO You need to invalidate the cache
        #      when the image data changes.
        self.__cache = cache.Cache(maxsize=1000)


    def makeLabel(self):
        """Returns a string representation of this ``VoxelDataSeries``
        instance.
        """

        display = self.displayCtx.getDisplay(self.overlay)
        opts    = display.opts
        coords  = opts.getVoxel()

        if coords is not None:
            return '{} [{} {} {}]'.format(display.name,
                                          coords[0],
                                          coords[1],
                                          coords[2])
        else:
            return '{} [out of bounds]'.format(display.name)


    def getData(self):
        """Returns the ``(xdata, ydata)`` at the current voxel location.

        This method may be overridden by sub-classes.
        """

        xdata = None
        ydata = self.dataAtCurrentVoxel()

        if ydata is not None:
            xdata = np.arange(len(ydata))

        return xdata, ydata


    # The PlotCanvas uses a new thread to access
    # data every time the displaycontext location
    # changes. So we mark this method as mutually
    # exclusive to prevent multiple
    # near-simultaneous accesses to the same voxel
    # location. The first time that a voxel location
    # is accessed, its data is cached. So when
    # subsequent (blocked) accesses execute, they
    # will hit the cache instead of hitting the disk
    # (which is a good thing).
    @idle.mutex
    def dataAtCurrentVoxel(self):
        """Returns the data for the current voxel of the overlay.  This method
        is intended to be used within the :meth:`DataSeries.getData` method
        of sub-classes.

        An internal cache is used to avoid the need to retrieve data for the
        same voxel multiple times, as retrieving data from large compressed
        4D images can be time consuming.

        The location for the current voxel is calculated by the
        :meth:`currentVoxelLocation` method, and the data lookup is performed
        by the :meth:`currentVoxelData` method. These methods may be
        overridden by sub-classes.

        :returns: A ``numpy`` array containing the data at the current
                  voxel, or ``None`` if the current location is out of bounds
                  of the image.
        """

        location = self.currentVoxelLocation()

        if location is None:
            return None

        data = self.__cache.get(location, None)

        if data is None:
            data = self.currentVoxelData(location)
            self.__cache.put(location, data)

        return data


    def currentVoxelLocation(self):
        """Used by :meth:`dataAtCurrentVoxel`. Returns the current voxel
        location. This is used as a key for the voxel data cache implemented
        within the :meth:`dataAtCurrentVoxel` method, and subsequently passed
        to the :meth:`currentVoxelData` method.

        This method may be overridden by sub-classes.
        """

        opts  = self.displayCtx.getOpts(self.overlay)
        vdim  = opts.volumeDim
        voxel = opts.getVoxel()

        if voxel is None:
            return None

        x, y, z = voxel

        return (x, y, z, vdim)


    def currentVoxelData(self, location):
        """Used by :meth:`dataAtCurrentVoxel`. Returns the data at the
        specified location.

        This method may be overridden by sub-classes.
        """
        voxel = location[:3]
        opts  = self.displayCtx.getOpts(self.overlay)
        data  = self.overlay[opts.index(voxel, atVolume=False)]

        return data



#!/usr/bin/env python
#
# loadplugin.py - The LoadPluginAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LoadPluginAction` class, an action which
allows the user to load/install FSLeyes plugins.
"""


import os.path as op
import            os

import wx

import fsl.utils.settings           as fslsettings
import fsleyes_widgets.utils.status as status
import fsleyes.plugins              as plugins
import fsleyes.strings              as strings
from . import                          base


class LoadPluginAction(base.Action):
    """The :class:`LoadPluginAction` class is an :class:`.Action` which allows
    the user to load/install FSLeyes plugins - see the :mod:`.plugins` module.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``LoadPluginAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The top-level :class:`.DisplayContext`.
        :arg overlayList: The :class:`.FSLeyesFrame`.
        """
        base.Action.__init__(self, overlayList, displayCtx, self.__loadPlugin)
        self.__frame = frame


    def __loadPlugin(self, *args, **kwargs):
        """Prompts the user to select a plugin file, asks them whether they
        would like to install it permanently, and then passes it to either
        :func:`.loadPlugin` or :func:`.installPlugin`.
        """

        lastDir = fslsettings.read('loadPluginLastDir')

        if lastDir is None:
            lastDir = os.getcwd()

        msg = strings.messages[self, 'loadPlugin']
        dlg = wx.FileDialog(self.__frame,
                            message=msg,
                            defaultDir=lastDir,
                            wildcard='*.py',
                            style=wx.FD_OPEN)

        if dlg.ShowModal() != wx.ID_OK:
            return

        fname = dlg.GetPath()

        fslsettings.write('loadPluginLastDir', op.dirname(fname))

        dlg = wx.MessageDialog(
            self.__frame,
            caption=strings.titles[  self, 'installPlugin'],
            message=strings.messages[self, 'installPlugin'],
            style=wx.YES_NO | wx.CANCEL)

        result = dlg.ShowModal()

        if result == wx.ID_YES:
            etitle = strings.titles[  self, 'installError']
            emsg   = strings.messages[self, 'installError']
            func   = plugins.installPlugin
        elif result == wx.ID_NO:
            etitle = strings.titles[  self, 'loadError']
            emsg   = strings.messages[self, 'loadError']
            func   = plugins.loadPlugin
        else:
            return

        with status.reportIfError(title=etitle, msg=emsg, raiseError=False):
            func(fname)

        for panel in self.__frame.viewPanels:
            panel.reloadPlugins()

        self.__frame.refreshViewMenu()
        self.__frame.refreshToolsMenu()
        self.__frame.refreshSettingsMenu()



#!/usr/bin/env python
#
# lightboxtoolbar.py - The LightBoxToolBar class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LightBoxToolBar` class, which is a
:class:`.ControlToolBar` for use with the :class:`.LightBoxPanel`.
"""


import wx

import fsleyes_props                 as props

import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.views.lightboxpanel   as lightboxpanel
import fsleyes.toolbar               as fsltoolbar
import fsleyes.actions               as actions
import fsleyes.icons                 as fslicons
import fsleyes.tooltips              as fsltooltips
import fsleyes.strings               as strings


class LightBoxToolBar(ctrlpanel.ControlToolBar):
    """The ``LightBoxToolBar`` is a :class:`.ControlToolBar` for use with the
    :class:`.LightBoxPanel`. A ``LightBoxToolBar`` looks something like this:


    .. image:: images/lightboxtoolbar.png
       :scale: 50%
       :align: center


    The ``LightBoxToolBar`` allows the user to control important parts of the
    :class:`.LightBoxPanel` display, and also to display a
    :class:`.CanvasSettingsPanel`, which allows control over all aspects of a
    ``LightBoxPanel``.
    """


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``LightBoxToolBar`` is only intended to be added to
        :class:`.LightBoxPanel` views.
        """

        return [lightboxpanel.LightBoxPanel]


    def __init__(self, parent, overlayList, displayCtx, lb):
        """Create a ``LightBoxToolBar``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg lb:          The :class:`.LightBoxPanel` instance.
        """

        ctrlpanel.ControlToolBar.__init__(self,
                                          parent,
                                          overlayList,
                                          displayCtx,
                                          lb,
                                          height=24,
                                          kbFocus=True)

        self.lightBoxPanel = lb

        lbOpts = lb.sceneOpts

        coronalIcon          = 'coronalSlice24'
        coronalHighlightIcon = 'coronalSliceHighlight24'

        icons = {
            'screenshot'                : fslicons.findImageFile('camera24'),
            'movieMode'                 : [
                fslicons.findImageFile('movieHighlight24'),
                fslicons.findImageFile('movie24')],
            'CanvasSettingsPanel' : [
                fslicons.findImageFile('spannerHighlight24'),
                fslicons.findImageFile('spanner24')],

            'zax' : {
                0 : [fslicons.findImageFile('sagittalSliceHighlight24'),
                     fslicons.findImageFile('sagittalSlice24')],
                1 : [fslicons.findImageFile(coronalHighlightIcon),
                     fslicons.findImageFile(coronalIcon)],
                2 : [fslicons.findImageFile('axialSliceHighlight24'),
                     fslicons.findImageFile('axialSlice24')],
            }
        }

        tooltips = {

            'screenshot'   : fsltooltips.actions[   lb,      'screenshot'],
            'movieMode'    : fsltooltips.properties[lb,      'movieMode'],
            'zax'          : fsltooltips.properties[lbOpts,  'zax'],
            'sliceSpacing' : fsltooltips.properties[lbOpts,  'sliceSpacing'],
            'zrange'       : fsltooltips.properties[lbOpts,  'zrange'],
            'zoom'         : fsltooltips.properties[lbOpts,  'zoom'],
            'CanvasSettingsPanel' : fsltooltips.actions[
                lb, 'CanvasSettingsPanel'],
        }

        specs = {

            'CanvasSettingsPanel' : actions.ToggleActionButton(
                'CanvasSettingsPanel',
                actionKwargs={'floatPane' : True},
                icon=icons['CanvasSettingsPanel'],
                tooltip=tooltips['CanvasSettingsPanel']),

            'screenshot' : actions.ActionButton(
                'screenshot',
                icon=icons['screenshot'],
                tooltip=tooltips['screenshot']),

            'movieMode'    : props.Widget(
                'movieMode',
                icon=icons['movieMode'],
                tooltip=tooltips['movieMode']),

            'zax'          : props.Widget(
                'zax',
                icons=icons['zax'],
                tooltip=tooltips['zax']),

            'sliceSpacing' : props.Widget(
                'sliceSpacing',
                spin=False,
                showLimits=False,
                tooltip=tooltips['sliceSpacing']),

            'zrange'       : props.Widget(
                'zrange',
                spin=False,
                showLimits=False,
                tooltip=tooltips['zrange'],
                labels=[strings.choices[lbOpts, 'zrange', 'min'],
                        strings.choices[lbOpts, 'zrange', 'max']]),

            'zoom'         : props.Widget(
                'zoom',
                spin=False,
                showLimits=False,
                tooltip=tooltips['zoom']),
        }

        # Slice spacing and zoom go on a single panel
        panel = wx.Panel(self)
        sizer = wx.FlexGridSizer(2, 2, 0, 0)
        panel.SetSizer(sizer)

        more         = props.buildGUI(self,
                                      lb,
                                      specs['CanvasSettingsPanel'])
        screenshot   = props.buildGUI(self,  lb,         specs['screenshot'])
        movieMode    = props.buildGUI(self,  lb,         specs['movieMode'])
        zax          = props.buildGUI(self,  lbOpts,     specs['zax'])
        zrange       = props.buildGUI(self,  lbOpts,     specs['zrange'])
        zoom         = props.buildGUI(panel, lbOpts,     specs['zoom'])
        spacing      = props.buildGUI(panel, lbOpts,     specs['sliceSpacing'])
        zoomLabel    = wx.StaticText(panel)
        spacingLabel = wx.StaticText(panel)

        zoomLabel   .SetLabel(strings.properties[lbOpts, 'zoom'])
        spacingLabel.SetLabel(strings.properties[lbOpts, 'sliceSpacing'])

        sizer.Add(zoomLabel)
        sizer.Add(zoom,    flag=wx.EXPAND)
        sizer.Add(spacingLabel)
        sizer.Add(spacing, flag=wx.EXPAND)

        tools = [more,
                 screenshot,
                 fsltoolbar.ToolBarDivider(self,
                                           height=24,
                                           orient=wx.VERTICAL),
                 zax,
                 fsltoolbar.ToolBarDivider(self,
                                           height=24,
                                           orient=wx.VERTICAL),
                 movieMode,
                 zrange,
                 panel]
        nav   = [more, screenshot, zax, movieMode, zrange, zoom, spacing]

        self.SetTools(tools)
        self.setNavOrder(nav)
